#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################
import time
import threading

import numpy

from unittest import TestCase
from taurus.test import insertTest

from sardana.sardanautils import is_number, is_pure_str
from sardana.pool import AcqSynch, AcqMode
from sardana.pool.pooldefs import SynchDomain, SynchParam
from sardana.pool.poolsynchronization import PoolSynchronization
from sardana.pool.poolacquisition import PoolAcquisitionHardware, \
    PoolAcquisitionSoftware, PoolAcquisitionSoftwareStart, \
    get_acq_ctrls, get_timerable_ctrls
from sardana.sardanathreadpool import get_thread_pool
from sardana.pool.test import createControllerConfiguration, \
    createTimerableControllerConfiguration, BasePoolTestCase, FakeElement, \
    AttributeListener


class AcquisitionTestCase(BasePoolTestCase):

    CHANNEL_NAME = None  # Override this!
    SYNCHRONIZATION = None  # Override this!

    def setUp(self):
        """Create dummy controllers and elements."""
        BasePoolTestCase.setUp(self)
        self.acquisition = None
        self.synchronization = None
        self.data_listener = AttributeListener()
        self.main_element = FakeElement(self.pool)
        self.tg = self.tgs['_test_tg_1_1']
        self.tg_ctrl = self.tg.get_controller()
        self.channel = self.exp_channels[self.CHANNEL_NAME]
        self.channel_ctrl = self.channel.get_controller()
        self.channel_names = [self.CHANNEL_NAME]

    def create_action(self, class_, elements):
        action = class_(self.main_element)
        for element in elements:
            action.add_element(element)
        return action

    def add_listeners(self, chn_list):
        for chn in chn_list:
            chn.add_listener(self.data_listener)

    def _prepare(self, integ_time, repetitions, latency_time, nb_starts):
        pass

    def prepare(self, integ_time, repetitions, latency_time, nb_starts):
        self.channel_ctrl.set_ctrl_par("synchronization",
                                       self.SYNCHRONIZATION)
        self._prepare(integ_time, repetitions, latency_time, nb_starts)
        self.channel_ctrl.ctrl.PrepareOne(self.channel.axis, integ_time,
                                          repetitions, latency_time,
                                          nb_starts)

    def wait_finish(self):
        # waiting for acquisition and synchronization to finish
        while (self.acquisition._is_busy()
               or self.synchronization.is_running()):
            time.sleep(.1)

    def do_asserts(self, repetitions, jobs_before, strict=True):
        table = self.data_listener.get_table()
        header = table.dtype.names
        # checking if all channels produced data
        for channel in self.channel_names:
            msg = 'data from channel %s were not acquired' % channel
            self.assertIn(channel, header, msg)
        # checking if data were acquired
        for ch_name in header:
            ch_data_len = len(table[ch_name])
            if strict:
                msg = ('length of data for channel %s is %d and should be '
                       '%d' % (ch_name, ch_data_len, repetitions))
                self.assertEqual(repetitions, ch_data_len, msg)
            else:
                msg = ('length of data for channel %s is %d and should <= '
                       '%d' % (ch_name, ch_data_len, repetitions))
                self.assertGreaterEqual(repetitions, ch_data_len, msg)
            for value in table[ch_name]:
                if (isinstance(value, numpy.ndarray)
                        or is_number(value)
                        or is_pure_str(value)):
                    break
            else:
                raise AssertionError('channel %s does not report any '
                                     'valid data')
        # checking if there are no pending jobs
        jobs_after = get_thread_pool().qsize
        msg = ('there are %d jobs pending to be done after the acquisition ' +
               '(before: %d)') % (jobs_after, jobs_before)
        self.assertEqual(jobs_before, jobs_after, msg)

    def tearDown(self):
        BasePoolTestCase.tearDown(self)
        self.data_listener = None
        self.channel_names = None
        self.main_element = None


@insertTest(helper_name='continuous_acquisition', offset=0, active_interval=0.1,
            passive_interval=0.2, repetitions=10, integ_time=0.2)
@insertTest(helper_name='continuous_acquisition', offset=0,
            active_interval=0.001, passive_interval=0.21, repetitions=10,
            integ_time=0.1)
@insertTest(helper_name='continuous_acquisition', offset=0,
            active_interval=0.001, passive_interval=0.15, repetitions=10,
            integ_time=0.01)
@insertTest(helper_name='continuous_acquisition', offset=0,
            active_interval=0.001, passive_interval=0.1, repetitions=10,
            integ_time=0.01)
class DummyAcquisitionTestCase(AcquisitionTestCase, TestCase):
    """Integration test of PoolSynchronization, PoolAcquisitionHardware and
    PoolAcquisitionSoftware actions. This test plays the role of the
    PoolAcquisition macro action (it aggregates the sub-actions and assign the
    elements to corresponding sub-actions) and the PoolMeasurementGroup (it
    configures the elements and controllers).
    """

    CHANNEL_NAME = "_test_ct_1_1"

    def setUp(self):
        """Create a Controller, TriggerGate and PoolSynchronization objects from
        dummy configurations.
        """
        TestCase.setUp(self)
        AcquisitionTestCase.setUp(self)

    def event_received(self, *args, **kwargs):
        """Executes a single software triggered acquisition."""
        _, type_, index = args
        name = type_.name
        if name == "active":
            if self.sw_acq._is_busy():
                # skipping acquisition cause the previous on is ongoing
                return
            else:
                self.sw_acq._set_busy()
                args = self.sw_acq_args
                kwargs = self.sw_acq_kwargs
                kwargs['index'] = index
                get_thread_pool().add(self.sw_acq.run,
                                      self.sw_acq._set_ready,
                                      *args,
                                      **kwargs)

    def continuous_acquisition(self, offset, active_interval, passive_interval,
                               repetitions, integ_time):
        """Executes measurement running the TGGeneration and Acquisition actions
        according the test parameters. Checks the lengths of the acquired data.
        """
        # obtaining elements created in the BasePoolTestCase.setUp
        tg_1_1 = self.tgs['_test_tg_1_1']
        tg_ctrl_1 = tg_1_1.get_controller()
        ct_1_1 = self.cts['_test_ct_1_1']  # hw synchronized
        ct_2_1 = self.cts['_test_ct_2_1']  # sw synchronized
        ct_ctrl_1 = ct_1_1.get_controller()
        ct_ctrl_1.set_ctrl_par("synchronization", AcqSynch.HardwareTrigger)
        ct_ctrl_2 = ct_2_1.get_controller()
        self.channel_names.append('_test_ct_1_1')
        self.channel_names.append('_test_ct_2_1')

        conf_ct_ctrl_1 = createTimerableControllerConfiguration(ct_ctrl_1,
                                                                [ct_1_1])
        conf_ct_ctrl_2 = createTimerableControllerConfiguration(ct_ctrl_2,
                                                                [ct_2_1])
        hw_ctrls = get_timerable_ctrls([conf_ct_ctrl_1],
                                       acq_mode=AcqMode.Timer)
        sw_ctrls = get_timerable_ctrls([conf_ct_ctrl_2],
                                       acq_mode=AcqMode.Timer)
        sw_master = sw_ctrls[0].master
        conf_tg_ctrl_1 = createControllerConfiguration(tg_ctrl_1, [tg_1_1])
        synch_ctrls = get_acq_ctrls([conf_tg_ctrl_1])
        # creating synchronization action
        self.synchronization = self.create_action(PoolSynchronization,
                                                  [tg_1_1])
        self.synchronization.add_listener(self)
        # add_listeners
        self.add_listeners([ct_1_1, ct_2_1])
        # creating acquisition actions
        self.hw_acq = self.create_action(PoolAcquisitionHardware, [ct_1_1])
        self.sw_acq = self.create_action(PoolAcquisitionSoftware, [ct_2_1])
        self.sw_acq_args = (sw_ctrls, integ_time, sw_master)
        self.sw_acq_kwargs = dict(synch=True)

        total_interval = active_interval + passive_interval
        group = {
            SynchParam.Delay: {SynchDomain.Time: offset},
            SynchParam.Active: {SynchDomain.Time: active_interval},
            SynchParam.Total: {SynchDomain.Time: total_interval},
            SynchParam.Repeats: repetitions
        }
        synch_description = [group]
        # get the current number of jobs
        jobs_before = get_thread_pool().qsize
        self.hw_acq.run(hw_ctrls, integ_time, repetitions, 0)
        self.synchronization.run(synch_ctrls, synch_description)
        # waiting for acquisition and synchronization to finish
        while (self.hw_acq.is_running()
               or self.sw_acq._is_busy()
               or self.synchronization.is_running()):
            time.sleep(.1)
        self.do_asserts(repetitions, jobs_before)

    def tearDown(self):
        AcquisitionTestCase.tearDown(self)
        TestCase.tearDown(self)


class BaseAcquisitionSoftwareTestCase(AcquisitionTestCase):
    """Base class for integration tests of PoolSynchronization and
    PoolAcquisitionSoftware"""

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        AcquisitionTestCase.setUp(self)

    def event_received(self, *args, **kwargs):
        """Callback to execute software start acquisition."""
        _, type_, value = args
        name = type_.name
        if name == "active":
            if self.acquisition._is_busy():
                # skipping acquisition cause the previous on is ongoing
                return
            else:
                self.acquisition._set_busy()
                acq_args = list(self.acq_args)
                acq_kwargs = self.acq_kwargs
                index = value
                acq_args[3] = index
                get_thread_pool().add(self.acquisition.run,
                                      self.acquisition._set_ready,
                                      *acq_args,
                                      **acq_kwargs)

    def acquire(self, integ_time, repetitions, latency_time):
        """Acquire with a dummy C/T synchronized by a hardware start
        trigger from a dummy T/G."""
        self.prepare(integ_time, repetitions, latency_time, 1)
        conf_ct_ctrl_1 = createTimerableControllerConfiguration(
            self.channel_ctrl, [self.channel])
        ctrls = get_timerable_ctrls([conf_ct_ctrl_1], AcqMode.Timer)
        master = ctrls[0].master
        # creating synchronization action
        self.synchronization = self.create_action(PoolSynchronization,
                                                  [self.tg])
        self.synchronization.add_listener(self)
        # add_listeners
        self.add_listeners([self.channel])
        # creating acquisition actions
        self.acquisition = self.create_action(PoolAcquisitionSoftware,
                                              [self.channel])
        self.acq_args = (ctrls, integ_time, master, None)
        self.acq_kwargs = dict(synch=True)

        total_interval = integ_time + latency_time
        group = {
            SynchParam.Delay: {SynchDomain.Time: 0},
            SynchParam.Active: {SynchDomain.Time: integ_time},
            SynchParam.Total: {SynchDomain.Time: total_interval},
            SynchParam.Repeats: repetitions
        }
        synch_description = [group]
        # get the current number of jobs
        jobs_before = get_thread_pool().qsize
        self.synchronization.run([], synch_description)
        self.wait_finish()
        self.do_asserts(repetitions, jobs_before, strict=False)

    def tearDown(self):
        AcquisitionTestCase.tearDown(self)
        TestCase.tearDown(self)


class BaseAcquisitionSoftwareStartTestCase(AcquisitionTestCase):
    """Base class for integration tests of PoolSynchronization and
    PoolAcquisitionSoftwareStart"""

    SYNCHRONIZATION = AcqSynch.SoftwareStart

    def setUp(self):
        """Create test actors (controllers and elements)"""
        AcquisitionTestCase.setUp(self)

    def event_received(self, *args, **kwargs):
        """Callback to execute software start acquisition."""
        _, type_, value = args
        name = type_.name
        if name == "start":
            self.acquisition._set_busy()
            get_thread_pool().add(self.acquisition.run,
                                  self.acquisition._set_ready,
                                  *self.acq_args,
                                  **self.acq_kwargs)

    def acquire(self, integ_time, repetitions, latency_time):
        """Acquire with a dummy C/T synchronized by a hardware start
        trigger from a dummy T/G."""
        self.prepare(integ_time, repetitions, latency_time, 1)
        conf_ct_ctrl_1 = createTimerableControllerConfiguration(
            self.channel_ctrl, [self.channel])
        ctrls = get_timerable_ctrls([conf_ct_ctrl_1], AcqMode.Timer)
        master = ctrls[0].master
        # creating synchronization action
        self.synchronization = self.create_action(PoolSynchronization,
                                                  [self.tg])
        self.synchronization.add_listener(self)
        # add_listeners
        self.add_listeners([self.channel])
        # creating acquisition actions
        self.acquisition = self.create_action(PoolAcquisitionSoftwareStart,
                                              [self.channel])
        self.acq_args = (ctrls, integ_time, master, repetitions, latency_time)
        self.acq_kwargs = dict(synch=True)

        total_interval = integ_time + latency_time
        group = {
            SynchParam.Delay: {SynchDomain.Time: 0},
            SynchParam.Active: {SynchDomain.Time: integ_time},
            SynchParam.Total: {SynchDomain.Time: total_interval},
            SynchParam.Repeats: repetitions
        }
        synch_description = [group]
        # get the current number of jobs
        jobs_before = get_thread_pool().qsize
        self.synchronization.run([], synch_description)
        self.wait_finish()
        self.do_asserts(repetitions, jobs_before, strict=False)

    def tearDown(self):
        AcquisitionTestCase.tearDown(self)


class BaseAcquisitionHardwareTestCase(AcquisitionTestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware"""

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        AcquisitionTestCase.setUp(self)

    def acquire(self, integ_time, repetitions, latency_time):
        """Acquire with a dummy C/T synchronized by a hardware start
        trigger from a dummy T/G."""
        self.prepare(integ_time, repetitions, latency_time, 1)
        conf_ct_ctrl_1 = createTimerableControllerConfiguration(
            self.channel_ctrl, [self.channel])
        ctrls = get_timerable_ctrls([conf_ct_ctrl_1], AcqMode.Timer)
        conf_tg_ctrl_1 = createControllerConfiguration(self.tg_ctrl,
                                                       [self.tg])
        synch_ctrls = get_acq_ctrls([conf_tg_ctrl_1])
        self.synchronization = self.create_action(PoolSynchronization,
                                                  [self.tg])
        # add data listeners
        self.add_listeners([self.channel])
        # creating acquisition actions
        self.acquisition = self.create_action(PoolAcquisitionHardware,
                                              [self.channel])
        self.acq_args = ([conf_ct_ctrl_1], integ_time, repetitions)
        # prepare synchronization description
        total_interval = integ_time + latency_time
        group = {
            SynchParam.Delay: {SynchDomain.Time: 0},
            SynchParam.Active: {SynchDomain.Time: integ_time},
            SynchParam.Total: {SynchDomain.Time: total_interval},
            SynchParam.Repeats: repetitions
        }
        synch_description = [group]
        # get the current number of jobs
        jobs_before = get_thread_pool().qsize
        self.acquisition.run(ctrls, integ_time, repetitions, 0)
        self.synchronization.run(synch_ctrls, synch_description)
        self.wait_finish()
        self.do_asserts(repetitions, jobs_before)

    def wait_finish(self):
        # waiting for acquisition and synchronization to finish
        while (self.acquisition.is_running()
               or self.synchronization.is_running()):
            time.sleep(.1)

    def tearDown(self):
        AcquisitionTestCase.tearDown(self)
        TestCase.tearDown(self)


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class AcquisitionCTSoftwareTriggerTestCase(BaseAcquisitionSoftwareTestCase,
                                           TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionSoftware
    with dummy CT channel synchronized by software trigger."""

    CHANNEL_NAME = "_test_ct_1_1"
    SYNCHRONIZATION = AcqSynch.SoftwareTrigger


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class Acquisition2DSoftwareTriggerTestCase(BaseAcquisitionSoftwareTestCase,
                                           TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionSoftware
    with dummy 2D channel synchronized by software trigger."""

    CHANNEL_NAME = "_test_2d_1_1"
    SYNCHRONIZATION = AcqSynch.SoftwareTrigger

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionSoftwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.1)
class Acquisition2DSoftwareTriggerRefTestCase(BaseAcquisitionSoftwareTestCase,
                                              TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by software trigger and configured to
    report value reference.
    """
    CHANNEL_NAME = "_test_2d_1_1"
    SYNCHRONIZATION = AcqSynch.SoftwareTrigger

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionSoftwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuerefbuffer")

    def _prepare(self, integ_time, repetitions, latency_time, nb_starts):
        self.channel.value_ref_enabled = True
        axis = self.channel.axis
        self.channel_ctrl.set_axis_par(axis, "value_ref_enabled", True)


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.1)
class AcquisitionCTSoftwareStartTestCase(
        BaseAcquisitionSoftwareStartTestCase, TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy CT channel synchronized by software start.
    """

    CHANNEL_NAME = "_test_ct_1_1"

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionSoftwareStartTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.1)
class Acquisition2DSoftwareStartTestCase(
        BaseAcquisitionSoftwareStartTestCase, TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by software start.
    """

    CHANNEL_NAME = "_test_2d_1_1"

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionSoftwareStartTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")

    def _prepare(self, integ_time, repetitions, latency_time, nb_starts):
        axis = self.channel.axis
        self.channel_ctrl.set_axis_par(axis, "value_ref_enabled", False)
        self.channel.value_ref_enabled = False


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.1)
class Acquisition2DSoftwareStartRefTestCase(
        BaseAcquisitionSoftwareStartTestCase, TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by software start and configured to
    report value reference.
    """
    CHANNEL_NAME = "_test_2d_1_1"

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionSoftwareStartTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuerefbuffer")

    def _prepare(self, integ_time, repetitions, latency_time, nb_starts):
        self.channel.value_ref_enabled = True
        axis = self.channel.axis
        self.channel_ctrl.set_axis_par(axis, "value_ref_enabled", True)


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class AcquisitionCTHardwareStartTestCase(BaseAcquisitionHardwareTestCase,
                                         TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy CT channel synchronized by hardware start.
    """
    CHANNEL_NAME = "_test_ct_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareStart

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        AcquisitionTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class Acquisition2DHardwareStartTestCase(BaseAcquisitionHardwareTestCase,
                                         TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by hardware start.
    """
    CHANNEL_NAME = "_test_2d_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareStart

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionHardwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class Acquisition2DHardwareStartRefTestCase(
        BaseAcquisitionHardwareTestCase, TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by hardware start and configured to
    report value reference.
    """
    CHANNEL_NAME = "_test_2d_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareStart

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionHardwareTestCase.setUp(self)
        self.channel_ctrl.set_log_level(10)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuerefbuffer")

    def _prepare(self, integ_time, repetitions, latency_time, nb_starts):
        self.channel.value_ref_enabled = True
        axis = self.channel.axis
        self.channel_ctrl.set_axis_par(axis, "value_ref_enabled", True)


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class AcquisitionCTHardwareTriggerTestCase(BaseAcquisitionHardwareTestCase,
                                           TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy CT channel synchronized by hardware trigger."""

    CHANNEL_NAME = "_test_ct_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareTrigger

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionHardwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class Acquisition2DHardwareTriggerTestCase(BaseAcquisitionHardwareTestCase,
                                           TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by hardware trigger."""

    CHANNEL_NAME = "_test_2d_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareTrigger

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionHardwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class Acquisition2DHardwareTriggerRefTestCase(BaseAcquisitionHardwareTestCase,
                                              TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by hardware trigger and configured to
    report value reference.
    """
    CHANNEL_NAME = "_test_2d_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareTrigger

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionHardwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuerefbuffer")

    def _prepare(self, integ_time, repetitions, latency_time, nb_starts):
        self.channel.value_ref_enabled = True
        axis = self.channel.axis
        self.channel_ctrl.set_axis_par(axis, "value_ref_enabled", True)


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class AcquisitionCTHardwareGateTestCase(BaseAcquisitionHardwareTestCase,
                                        TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy CT channel synchronized by hardware gate."""

    CHANNEL_NAME = "_test_ct_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareGate

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionHardwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class Acquisition2DHardwareGateTestCase(BaseAcquisitionHardwareTestCase,
                                        TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by hardware gate."""

    CHANNEL_NAME = "_test_2d_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareGate

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionHardwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuebuffer")


@insertTest(helper_name='acquire', integ_time=0.01, repetitions=10,
            latency_time=0.02)
class Acquisition2DHardwareGateRefTestCase(BaseAcquisitionHardwareTestCase,
                                           TestCase):
    """Integration test of PoolSynchronization and PoolAcquisitionHardware
    with dummy 2D channel synchronized by hardware gate and configured to
    report value reference.
    """
    CHANNEL_NAME = "_test_2d_1_1"
    SYNCHRONIZATION = AcqSynch.HardwareGate

    def setUp(self):
        """Create test actors (controllers and elements)"""
        TestCase.setUp(self)
        BaseAcquisitionHardwareTestCase.setUp(self)
        self.data_listener = AttributeListener(dtype=object,
                                               attr_name="valuerefbuffer")

    def _prepare(self, integ_time, repetitions, latency_time, nb_starts):
        self.channel.value_ref_enabled = True
        axis = self.channel.axis
        self.channel_ctrl.set_axis_par(axis, "value_ref_enabled", True)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import uuid
from unittest import TestCase

from tango import DevState
from taurus import Device
from taurus.test.base import insertTest

from .test_pool import is_numerical
from sardana.pool.pooldefs import AcqSynchType
from sardana.taurus.core.tango.sardana.pool import registerExtensions
from sardana.tango.pool.test.base_sartest import SarTestTestCase


@insertTest(helper_name="stress_count",
            test_method_doc="stress count with CT (hardware trigger) and 0D",
            elements=["_test_ct_1_1", "_test_0d_1_1"], repeats=100,
            synchronizer="_test_tg_1_1", synchronization=AcqSynchType.Trigger)
@insertTest(helper_name="stress_count",
            test_method_doc="stress count with CT (software trigger) and 0D",
            elements=["_test_ct_1_1", "_test_0d_1_1"], repeats=100,
            synchronizer="software", synchronization=AcqSynchType.Trigger)
@insertTest(helper_name="stress_count",
            test_method_doc="stress count with CT (hardware start)",
            elements=["_test_ct_1_1"], repeats=100,
            synchronizer="_test_tg_1_1", synchronization=AcqSynchType.Start)
@insertTest(helper_name="stress_count",
            test_method_doc="stress count with CT (software start)",
            elements=["_test_ct_1_1"], repeats=100,
            synchronizer="software", synchronization=AcqSynchType.Start)
@insertTest(helper_name="stress_count",
            test_method_doc="stress count with CT (hardware trigger)",
            elements=["_test_ct_1_1"], repeats=100,
            synchronizer="_test_tg_1_1", synchronization=AcqSynchType.Trigger)
@insertTest(helper_name="stress_count",
            test_method_doc="count with CT (software trigger)",
            elements=["_test_ct_1_1"], repeats=100,
            synchronizer="software", synchronization=AcqSynchType.Trigger)
class TestStressMeasurementGroup(SarTestTestCase, TestCase):

    def setUp(self):
        SarTestTestCase.setUp(self)
        registerExtensions()

    def stress_count(self, elements, repeats, synchronizer, synchronization):
        mg_name = str(uuid.uuid1())
        argin = [mg_name] + elements
        self.pool.CreateMeasurementGroup(argin)
        try:
            mg = Device(mg_name)
            mg.setSynchronizer(synchronizer, elements[0], apply=False)
            mg.setSynchronization(synchronization, elements[0])
            for i in range(repeats):
                state, values = mg.count(.001)
                self.assertEqual(state, DevState.ON,
                                 "wrong state after measurement {}".format(i))
                for channel_name, value in values.items():
                    msg = ("Value {} for {} is not numerical in "
                           "measurement {}").format(value, channel_name, i)
                    self.assertTrue(is_numerical(value), msg)
        finally:
            mg.cleanUp()
            self.pool.DeleteElement(mg_name)

    def tearDown(self):
        SarTestTestCase.tearDown(self)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the macro server scan data module"""

__all__ = ["ColumnDesc", "MoveableDesc", "Record", "RecordEnvironment",
           "ScanDataEnvironment", "RecordList", "ScanData", "ScanFactory"]

import copy
import math

from taurus.core.util.singleton import Singleton
from taurus import Device, Attribute, getSchemeFromName, Factory
from taurus.core.taurusexception import TaurusException
from taurus.core import TaurusElementType
from taurus import Release as taurus_release

from sardana.macroserver.scan.recorder import DataHandler


class ColumnDesc(object):
    """The description of a column for a Record"""

    _TYPE_MAP = {"short": "int16",
                 "ushort": "uint16"}
    _shape = []
    _dtype = 'float64'

    def __init__(self, **kwargs):
        """Expected keywords are:

               - name (str, mandatory): unique column name
               - label (str, optional): column label (defaults to name)
               - dtype (str, optional): data type. Defaults to 'float64'
               - shape (seq, optional): data shape. Defaults to []

        Any keyword not in the previous list will be converted to a member of
        the :class:`ColumnDesc`"""
        # enforce that the mandatory arguments are present
        try:
            self.name = kwargs.pop('name')
        except:
            raise TypeError('"name" parameter is mandatory')

        # make sure that at least the required members exist
        self.label = kwargs.pop('label', self.name)
        self.setDtype(kwargs.pop('dtype', self.__class__._dtype))
        self.setShape(kwargs.pop('shape', self.__class__._shape))

        # create members of the ColumnDesc class using the remaining keyword
        # args
        self._extra_kwargs = kwargs
        self.__dict__.update(kwargs)

    def getShape(self):
        return self._shape

    def setShape(self, shape):
        self._shape = self._simplifyShape(shape)

    def getDtype(self):
        return self._dtype

    def setDtype(self, dtype):
        self._dtype = self.tr(dtype)

    shape = property(getShape, setShape)
    dtype = property(getDtype, setDtype)

    @staticmethod
    def _simplifyShape(s):
        '''the idea is to strip the shape of useless "ones" at the beginning.
        For example:

            - () -> ()
            - (1,) -> ()
            - (1,1,1,1) -> ()
            - (2,) -> (2)
            - (1,2) -> (2)
            - (1,2,3) -> (2,3)
            - (2,3) -> (2,3)
            - (1,1,1,2,3) -> (2,3)
            - (3,1,1) -> (3,1,1)
        '''
        s = list(s)
        for i, e in enumerate(s):
            if e > 1:
                return s[i:]
        return []

    def tr(self, dtype):
        return self._TYPE_MAP.get(dtype, dtype)

    def toDict(self):
        '''creates a dictionary representation of the record'''
        d = copy.deepcopy(self._extra_kwargs)
        for k in ['name', 'label', 'dtype', 'shape']:
            d[k] = getattr(self, k)
        return d

    def clone(self):
        return copy.deepcopy(self)
        # return self.__class__(**self.toDict())


class MoveableDesc(ColumnDesc):

    def __init__(self, **kwargs):
        """Expected keywords are:

               - moveable (Moveable, mandatory): moveable object
               - name (str, optional): column name (defaults to moveable name)
               - label (str, optional): column label (defaults to moveable
                 name)
               - dtype (str, optional): data type. Defaults to 'float64'
               - shape (seq, optional): data shape. Defaults to (1,)
               - instrument (Instrument, optional): instrument object.
                 Defaults to moveable instrument"""

        try:
            self.moveable = moveable = kwargs.pop('moveable')
        except KeyError:
            raise TypeError("moveable parameter is mandatory")
        name = moveable.getName()
        kwargs['name'] = kwargs.get('name', name)
        kwargs['label'] = kwargs.get('label', name)
        kwargs['instrument'] = kwargs.get('instrument', moveable.instrument)

        self.min_value = kwargs.get('min_value')
        self.max_value = kwargs.get('max_value')
        self.is_reference = kwargs.get('is_reference')
        ColumnDesc.__init__(self, **kwargs)

    def toDict(self):
        d = ColumnDesc.toDict(self)
        d['min_value'] = self.min_value
        d['max_value'] = self.max_value
        d['is_reference'] = self.is_reference
        return d

    def clone(self):
        return self.__class__(moveable=self.moveable, **self.toDict())


class Record(object):
    """ One record is a set of values measured at the same time.

    The Record.data member will be
    a dictionary containing:
      - 'point_nb' : (int) the point number of the scan
      - for each column of the scan (motor or counter), a key with the
      corresponding column name (str) and the corresponding value
    """

    def __init__(self, data):
        self.recordno = 0
        self.data = data
        self.complete = 0
        self.written = 0

    def setRecordNo(self, recordno):
        self.recordno = recordno

    def setComplete(self):
        self.complete = 1

    def setWritten(self):
        self.written = 1

    def __get_t3_name(self, item):
        # check if the item is a device name or an attribute name
        try:
            proxy = Device(item)
        except TaurusException:
            try:
                proxy = Attribute(item)
            except TaurusException:
                raise KeyError(item)

        v = proxy.getNameValidator()
        params = v.getUriGroups(proxy.getFullName())
        name = '{0}:{1}/{2}'.format(params['host'].split('.')[0],
                                    params['port'],
                                    params['devname'])

        attr_name = params.get('_shortattrname', None)
        if attr_name is not None:
            name = '{0}/{1}'.format(name, attr_name)

        return name

    def __get_t4_name(self, item):
        scheme = getSchemeFromName(item)
        f = Factory(scheme=scheme)
        element_types = f.getValidTypesForName(item)
        if TaurusElementType.Attribute in element_types:
            validator = f.getAttributeNameValidator()
        elif TaurusElementType.Device in element_types:
            validator = f.getDeviceNameValidator()
        else:
            raise KeyError(item)

        name = validator.getNames(item)[0]
        return name

    def __get_tango_name(self, item):
        name = self.__get_t4_name(item)
        return name

    def __getitem__(self, item):
        item = item.lower()
        if item == 'dt':
            item = 'timestamp'
        if item in self.data:
            return self.data[item]

        # --------------------------------------------------------------------
        # TODO: refactor this block once data uses taurus 4 names

        if int(taurus_release.version.split('.')[0]) < 4:
            # Taurus 3 backward compatibility
            name = self.__get_t3_name(item)
            data = self.data[name]
        else:
            try:
                name = self.__get_t4_name(item)
                data = self.data[name]
            except KeyError:
                # Using a Tango URL
                name = self.__get_tango_name(item)
                data = self.data[name]
        # --------------------------------------------------------------------
        return data


class RecordEnvironment(dict):
    """  A RecordEnvironment is a set of arbitrary pairs of type
    label/value in the form of a dictionary.
    """
    __needed = ['title', 'labels']  # @TODO: it seems that this has changed
    # now labels are separated in moveables and counters

    def isValid(self):
        """ Check valid environment = all needed keys present """

        if not self.needed:
            return 1

        for ky in self.needed + self.__needed:
            if ky not in list(self.keys()):
                return 0
        else:
            return 1


class ScanDataEnvironment(RecordEnvironment):
    """It describes a recordlist and its environment

    A recordlist environment contains a number of predefined label/value pairs
    Values can be either a string, a numeric value or a list of strings,
    numbers

    title:     mandatory
    labels:    mandatory. label for each of the fields in a record of the
               recordlist
    fielddesc: description of the content of each of the fields in a record.
               Can be used to affect the way the field is saved by the
               recorder. If not present all fields are by default of type
               FLOAT and FORMAT ".8g"
    """
    needed = ['title', 'labels', 'user']


class RecordList(dict):
    """  A RecordList is a set of records: for example a scan.
    It is composed of a environment and a list of records"""

    def __init__(self, datahandler, environ=None, apply_interpolation=False,
                 apply_extrapolation=False, initial_data=None):

        self.datahandler = datahandler
        self.apply_interpolation = apply_interpolation
        self.apply_extrapolation = apply_extrapolation
        self.initial_data = initial_data
        if environ is None:
            self.environ = RecordEnvironment()
        else:
            self.environ = environ
        self.records = []
        # currentIndex indicates the place in the records list
        # where the next completed record will be written
        self.currentIndex = 0

    # make it pickable
    def __getstate__(self):
        return dict(datahandler=None, environ=None, records=self.records)

    def setEnviron(self, environ):
        self.environ = environ

    def updateEnviron(self, environ):
        self.environ.update(environ)

    def setEnvironValue(self, name, value):
        self.environ[name] = value

    def getEnvironValue(self, name):
        return self.environ[name]

    def getEnviron(self):
        return self.environ

    def start(self):
        self.recordno = 0
        # @TODO: it is necessary only by continuous scan
        # think how to separate this two cases
        self.columnIndexDict = {}
        self.labels = []
        self.refMoveablesLabels = []
        self.channelLabels = []
        self.currentIndex = 0
        self._mylabel = []

        for dataDesc in self.getEnvironValue('datadesc'):
            if isinstance(dataDesc, MoveableDesc):
                self.refMoveablesLabels.append(dataDesc.name)
            else:
                name = dataDesc.name
                if name not in ('point_nb', 'timestamp'):
                    self.channelLabels.append(name)
            self.labels.append(dataDesc.name)
        for label in self.labels:
            self.columnIndexDict[label] = 0
        ####
        self.datahandler.startRecordList(self)

    def initRecord(self):
        '''Init a dummy record and add it to the records list.
        A dummy record has:
           - point_nb of the consecutive record
           - each column initialized with NaN
           - each moveable initialized with None
        '''
        recordno = self.recordno
        if self.initial_data and recordno in self.initial_data:
            initial_data = self.initial_data.get(recordno)
        else:
            initial_data = dict()
        rc = Record({'point_nb': recordno})
        rc.data['timestamp'] = initial_data.get('timestamp')
        rc.setRecordNo(self.recordno)
        for label in self.channelLabels:
            rc.data[label] = initial_data.get(label, float('NaN'))
        for label in self.refMoveablesLabels:
            rc.data[label] = initial_data.get(label)
        self.records.append(rc)
        self.recordno += 1

    def initRecords(self, nb_records):
        '''Call nb_records times initRecord method
        '''
        for _ in range(nb_records):
            self.initRecord()

    def addRecord(self, record):
        rc = Record(record)
        rc.setRecordNo(self.recordno)
        self.records.append(rc)
        self[self.recordno] = rc
        self.recordno += 1
        self.datahandler.addRecord(self, rc)
        self.currentIndex += 1

    def applyZeroOrderInterpolation(self, record):
        ''' Apply a zero order interpolation to the given record
        '''
        if self.currentIndex > 0:
            data = record.data
            prev_data = self.records[self.currentIndex - 1].data
            for k, v in list(data.items()):
                if v is None:
                    continue
                # numpy arrays (1D or 2D) are valid values and does not require
                # interpolation but provokes TypeError
                try:
                    interpolate = math.isnan(v)
                except TypeError:
                    interpolate = False
                if interpolate:
                    data[k] = prev_data[k]

    def applyExtrapolation(self, record):
        """Apply extrapolation to the given record"""
        data = record.data
        for k, v in list(data.items()):
            if v is None:
                continue
            # numpy arrays (1D or 2D) are valid values and does not require
            # extrapolation but provokes TypeError
            try:
                extrapolate = math.isnan(v)
            except TypeError:
                extrapolate = False
            if extrapolate:
                next_idx = self.currentIndex + 1
                # dig into the record list for the first valid value in the
                # column and use it to extrapolate missing initial values
                while True:
                    next_data = self.records[next_idx].data
                    next_v = next_data[k]
                    try:
                        is_valid = not math.isnan(next_v)
                    except TypeError:
                        is_valid = True
                    if is_valid:
                        break
                    next_idx += 1
                data[k] = next_v

    def addData(self, data):
        """Adds data to the record list

        :param data: dictionary with two mandatory elements: label - string
                     and data - list of values
        :type data:  dict"""
        label = data['label']
        idxs = data['index']
        # TODO: think if the ScanData.addData is the best API for
        # passing value references
        rawData = data.get('value') or data.get('value_ref')


        maxIdx = max(idxs)
        recordsLen = len(self.records)
        # Calculate missing records
        missingRecords = recordsLen - (maxIdx + 1)
        # TODO: implement proper handling of timestamps and moveables
        if missingRecords < 0:
            missingRecords = abs(missingRecords)
            self.initRecords(missingRecords)
        for idx, value in zip(idxs, rawData):
            rc = self.records[idx]
            rc.setRecordNo(idx)
            rc.data[label] = value
            self.columnIndexDict[label] = idx + 1
        self.tryToAdd(idx, label)

    def tryToAdd(self, idx, label):
        start = self.currentIndex
        # apply extrapolation only at the beginning of the record list
        apply_extrapolation = (self.apply_extrapolation and start == 0)
        for i in range(start, idx + 1):
            if self.isRecordCompleted(i):
                rc = self.records[i]
                if apply_extrapolation:
                    self.applyExtrapolation(rc)
                self[self.currentIndex] = rc
                if self.apply_interpolation:
                    self.applyZeroOrderInterpolation(rc)
                self.datahandler.addRecord(self, rc)
                self.currentIndex += 1

    def isRecordCompleted(self, recordno):
        rc = self.records[recordno]
        for label in self.channelLabels:
            if self.columnIndexDict[label] <= self.currentIndex:
                return False
        rc.completed = 1
        return True

    def addRecords(self, records):
        list(map(self.addRecord, records))

    def end(self):
        start = self.currentIndex
        for i in range(start, len(self.records)):
            rc = self.records[i]
            self[self.currentIndex] = rc
            if self.apply_interpolation:
                self.applyZeroOrderInterpolation(rc)
            self.datahandler.addRecord(self, rc)
            self.currentIndex += 1
        self.datahandler.endRecordList(self)

    def getDataHandler(self):
        return self.datahandler


class ScanData(RecordList):

    def __init__(self, environment=None, data_handler=None,
                 apply_interpolation=False, apply_extrapolation=False):
        dh = data_handler or DataHandler()
        RecordList.__init__(self, dh, environment, apply_interpolation,
                            apply_extrapolation)


class ScanFactory(Singleton):

    def __init__(self):
        """ Initialization. Nothing to be done here for now."""
        pass

    def init(self, *args):
        """Singleton instance initialization."""
        pass

    def getDataHandler(self):
        return DataHandler()

    def getScanData(self, dh, apply_interpolation=False,
                    apply_extrapolation=False):
        return ScanData(data_handler=dh,
                        apply_interpolation=apply_interpolation,
                        apply_extrapolation=apply_extrapolation)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Tests for parser utilities."""

import unittest
from taurus.test import insertTest
from sardana.util.parser import ParamParser


pt0_params_def = []

pt1d_params_def = [
    {
        "default_value": 99,
        "description": "some bloody float",
        "max": None,
        "min": 1,
        "name": "value",
        "type": "Float"
    }
]

pt2_params_def = [
    {
        "default_value": None,
        "description": "some bloody motor",
        "max": None,
        "min": 1,
        "name": "motor",
        "type": "Motor"
    }
]
pt3_params_def = [
    {
        "default_value": None,
        "description": "List of values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "value",
                "max": None,
                "min": 1,
                "name": "position",
                "type": "Float"
            }
        ]
    }
]

pt3d_params_def = [
    {
        "default_value": None,
        "description": "List of values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": 21,
                "description": "value",
                "max": None,
                "min": 1,
                "name": "position",
                "type": "Float"
            }
        ]
    }
]

pt4_params_def = [
    {
        "default_value": None,
        "description": "List of motors",
        "max": None,
        "min": 1,
        "name": "motor_list",
        "type": [
            {
                "default_value": None,
                "description": "motor name",
                "max": None,
                "min": 1,
                "name": "motor",
                "type": "Motor"
            }
        ]
    }
]

pt5_params_def = [
    {
        "default_value": None,
        "description": "Motor to move",
        "max": None,
        "min": 1,
        "name": "motor",
        "type": "Motor"
    },
    {
        "default_value": None,
        "description": "List of values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "value",
                "max": None,
                "min": 1,
                "name": "pos",
                "type": "Float"
            }
        ]
    }
]

pt6_params_def = [
    {
        "default_value": None,
        "description": "Motor to move",
        "max": None,
        "min": 1,
        "name": "motor",
        "type": "Motor"
    },
    {
        "default_value": None,
        "description": "List of values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "value",
                "max": None,
                "min": 1,
                "name": "pos",
                "type": "Float"
            }
        ]
    }
]

pt7_params_def = [
    {
        "default_value": None,
        "description": "List of motor/position pairs",
        "max": None,
        "min": 1,
        "name": "m_p_pair",
        "type": [
            {
                "default_value": None,
                "description": "Motor to move",
                "max": None,
                "min": 1,
                "name": "motor",
                "type": "Motor"
            },
            {
                "default_value": None,
                "description": "Position to move to",
                "max": None,
                "min": 1,
                "name": "position",
                "type": "Float"
            }
        ]
    }
]

pt7d1_params_def = [
    {
        "default_value": None,
        "description": "List of motor/position pairs",
        "max": None,
        "min": 1,
        "name": "m_p_pair",
        "type": [
            {
                "default_value": None,
                "description": "Motor to move",
                "max": None,
                "min": 1,
                "name": "motor",
                "type": "Motor"
            },
            {
                "default_value": 2,
                "description": "Position to move to",
                "max": None,
                "min": 1,
                "name": "pos",
                "type": "Float"
            }
        ]
    }
]

pt7d2_params_def = [
    {
        "default_value": None,
        "description": "List of motor/position pairs",
        "max": None,
        "min": 1,
        "name": "m_p_pair",
        "type": [
            {
                "default_value": 'mot1',
                "description": "Motor to move",
                "max": None,
                "min": 1,
                "name": "motor",
                "type": "Motor"
            },
            {
                "default_value": 2,
                "description": "Position to move to",
                "max": None,
                "min": 1,
                "name": "pos",
                "type": "Float"
            }
        ]
    }
]

pt8_params_def = [
    {
        "default_value": None,
        "description": "List of motor/position pairs",
        "max": None,
        "min": 1,
        "name": "m_p_pair",
        "type": [
            {
                "default_value": None,
                "description": "Motor to move",
                "max": None,
                "min": 1,
                "name": "motor",
                "type": "Motor"
            },
            {
                "default_value": None,
                "description": "Position to move to",
                "max": 2,
                "min": 1,
                "name": "pos",
                "type": "Float"
            }
        ]
    }
]

pt9_params_def = [
    {
        "default_value": None,
        "description": "List of motor/position pairs",
        "max": None,
        "min": 1,
        "name": "m_p_pair",
        "type": [
            {
                "default_value": None,
                "description": "Motor to move",
                "max": None,
                "min": 1,
                "name": "motor",
                "type": "Motor"
            },
            {
                "default_value": None,
                "description": "Position to move to",
                "max": 2,
                "min": 1,
                "name": "pos",
                "type": "Float"
            }
        ]
    }
]

pt10_params_def = [
    {
        "default_value": None,
        "description": "List of values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "value",
                "max": None,
                "min": 1,
                "name": "pos",
                "type": "Float"
            }
        ]
    },
    {
        "default_value": None,
        "description": "Motor to move",
        "max": None,
        "min": 1,
        "name": "motor",
        "type": "Motor"
    },
]

pt11_params_def = [

    {
        "default_value": None,
        "description": "Counter to count",
        "max": None,
        "min": 1,
        "name": "counter",
        "type": "ExpChannel"
    },
    {
        "default_value": None,
        "description": "List of values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "value",
                "max": None,
                "min": 1,
                "name": "pos",
                "type": "Float"
            },
        ]
    },
    {
        "default_value": None,
        "description": "Motor to move",
        "max": None,
        "min": 1,
        "name": "motor",
        "type": "Motor"
    },
]

pt12_params_def = [

    {
        "default_value": None,
        "description": "List of values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "value",
                "max": None,
                "min": 1,
                "name": "pos",
                "type": "Float"
            },
        ]
    },
    {
        "default_value": None,
        "description": "List of Motors",
        "max": None,
        "min": 1,
        "name": "motor_list",
        "type": [
            {
                "default_value": 'mot1',
                "description": "Motor to move",
                "max": None,
                "min": 1,
                "name": "motor",
                "type": "Motor"
            },
        ]
    },
]

pt13_params_def = [
    {
        "default_value": None,
        "description": "Motor groups",
        "max": None,
        "min": 1,
        "name": "motor_group_list",
        "type": [
            {
                "default_value": None,
                "description": "List of motors",
                "max": None,
                "min": 1,
                "name": "motor list",
                "type": [
                    {
                        "default_value": None,
                        "description": "Motor to move",
                        "max": None,
                        "min": 1,
                        "name": "motor",
                        "type": "Motor"
                    }
                ]
            }
        ]
    }
]

pt14_params_def = [
    {
        "default_value": None,
        "description": "Motor groups",
        "max": None,
        "min": 1,
        "name": "motor_group_list",
        "type": [
            {
                "default_value": None,
                "description": "List of motors",
                "max": None,
                "min": 1,
                "name": "motor_list",
                "type": [
                    {
                        "default_value": None,
                        "description": "Motor to move",
                        "max": None,
                        "min": 1,
                        "name": "motor",
                        "type": "Motor"
                    }
                ]
            },
            {
                "default_value": None,
                "description": "Number",
                "max": None,
                "min": 1,
                "name": "float",
                "type": "Float"
            }
        ]
    }
]

pt14d_params_def = [
    {
        "default_value": None,
        "description": "Motor groups",
        "max": None,
        "min": 1,
        "name": "motor_group_list",
        "type": [
            {
                "default_value": None,
                "description": "List of motors",
                "max": None,
                "min": 1,
                "name": "motor_list",
                "type": [
                    {
                        "default_value": "mot1",
                        "description": "Motor to move",
                        "max": None,
                        "min": 1,
                        "name": "motor",
                        "type": "Motor"
                    }
                ]
            },
            {
                "default_value": 33,
                "description": "Number",
                "max": None,
                "min": 1,
                "name": "float",
                "type": "Float"
            }
        ]
    }
]

extra1_params_def = [
    {
        "default_value": None,
        "description": "Parameter",
        "max": None,
        "min": 1,
        "name": "param",
        "type": "String"
    },
    {
        "default_value": None,
        "description": "List of Scan files",
        "max": None,
        "min": 1,
        "name": "ScanFiles List",
        "type": [
            {
                "default_value": None,
                "description": "ScanFile",
                "max": None,
                "min": 1,
                "name": "ScanFile",
                "type": "String",
            }
        ]
    }
]

extra2_params_def = [

    {
        "default_value": None,
        "description": "Value 1",
        "max": None,
        "min": 1,
        "name": "value1",
        "type": "Float"
    },
    {
        "default_value": None,
        "description": "Value 2",
        "max": None,
        "min": 1,
        "name": "value2",
        "type": "float"
    },
    {
        "default_value": None,
        "description": "List of Strings",
        "max": None,
        "min": 1,
        "name": "string_list",
        "type": [
            {
                "default_value": None,
                "description": "string",
                "max": None,
                "min": 1,
                "name": "string",
                "type": "String"
            },
        ]
    },
]
extra3_params_def = [

    {
        "default_value": None,
        "description": "param",
        "max": None,
        "min": 1,
        "name": "param",
        "type": "String"
    },
    {
        "default_value": None,
        "description": "Value",
        "max": None,
        "min": 1,
        "name": "value",
        "type": "String"
    },
]

extra4_params_def = [

    {
        "default_value": None,
        "description": "value 1",
        "max": None,
        "min": 1,
        "name": "value1",
        "type": "Float"
    },
    {
        "default_value": None,
        "description": "Value 2",
        "max": None,
        "min": 1,
        "name": "value2",
        "type": "Float"
    },
]


extra5_params_def = [

    {
        "default_value": None,
        "description": "List of Motor and Values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "Motor",
                "max": None,
                "min": 1,
                "name": "pos",
                "type": "Motor"
            },
            {
                "default_value": None,
                "description": "Position to move to",
                "max": 2,
                "min": 1,
                "name": "pos",
                "type": "Float"
            }
        ]
    },
    {
        "default_value": None,
        "description": "Counter to use",
        "max": None,
        "min": 1,
        "name": "counter",
        "type": "ExpChan"
    },
    {
        "default_value": None,
        "description": "Value",
        "max": None,
        "min": 1,
        "name": "Value",
        "type": "Float"
    }
]
extra6_params_def = [

    {
        "default_value": None,
        "description": "List of Values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "Value 1",
                "max": None,
                "min": 1,
                "name": "value1",
                "type": "Float"
            },
            {
                "default_value": None,
                "description": "Value 2",
                "max": None,
                "min": 1,
                "name": "value2",
                "type": "Float"
            }
        ]
    }
]

extra7_params_def = [

    {
        "default_value": None,
        "description": "value 1",
        "max": None,
        "min": 1,
        "name": "value1",
        "type": "Float"
    },
    {
        "default_value": None,
        "description": "List of Values",
        "max": None,
        "min": 1,
        "name": "numb_list",
        "type": [
            {
                "default_value": None,
                "description": "Value 2.1",
                "max": None,
                "min": 1,
                "name": "value21",
                "type": "Float"
            },
            {
                "default_value": None,
                "description": "Value 2.2",
                "max": None,
                "min": 1,
                "name": "value22",
                "type": "Float"
            }
        ]
    }
]
extra8_params_def = [

    {
        "default_value": None,
        "description": "value 1",
        "max": None,
        "min": 1,
        "name": "value1",
        "type": "Float"
    },
    {
        "default_value": None,
        "description": "Value 2",
        "max": None,
        "min": 1,
        "name": "value2",
        "type": "Float"
    },
]


# parameters examples tests
@insertTest(helper_name="parse", params_def=pt0_params_def,
            params_str="", params=[])
@insertTest(helper_name="parse", params_def=pt1d_params_def,
            params_str="1", params=["1"])
@insertTest(helper_name="parse", params_def=pt1d_params_def,
            params_str="", params=[])
@insertTest(helper_name="parse", params_def=pt2_params_def,
            params_str="mot1", params=["mot1"])
@insertTest(helper_name="parse", params_def=pt3_params_def,
            params_str="1 34 15", params=[["1", "34", "15"]])
@insertTest(helper_name="parse", params_def=pt3_params_def,
            params_str="[1 34 15]", params=[["1", "34", "15"]])
@insertTest(helper_name="parse", params_def=pt3d_params_def,
            params_str="1 34 15", params=[["1", "34", "15"]])
@insertTest(helper_name="parse", params_def=pt3d_params_def,
            params_str="[1 34 15]", params=[["1", "34", "15"]])
@insertTest(helper_name="parse", params_def=pt3d_params_def,
            params_str="[1 [] 15]", params=[["1", [], "15"]])
@insertTest(helper_name="parse", params_def=pt4_params_def,
            params_str="[mot1 mot2 mot3]", params=[["mot1", "mot2", "mot3"]])
@insertTest(helper_name="parse", params_def=pt4_params_def,
            params_str="mot1 mot2 mot3", params=[["mot1", "mot2", "mot3"]])
@insertTest(helper_name="parse", params_def=pt5_params_def,
            params_str="mot1 1 3", params=["mot1", ["1", "3"]])
@insertTest(helper_name="parse", params_def=pt5_params_def,
            params_str="mot1 [1 3]", params=["mot1", ["1", "3"]])
@insertTest(helper_name="parse", params_def=pt6_params_def,
            params_str="mot1 [1 34 1]", params=["mot1", ["1", "34", "1"]])
@insertTest(helper_name="parse", params_def=pt6_params_def,
            params_str="mot1 1 34 1", params=["mot1", ["1", "34", "1"]])
@insertTest(helper_name="parse", params_def=pt7_params_def,
            params_str="mot1 1 mot2 3",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt7_params_def,
            params_str="[[mot1 1] [mot2 3]]",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt7d1_params_def,
            params_str="[[mot1 1] [mot2 3]]",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt7d1_params_def,
            params_str="mot1 1 mot2 3",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt7d1_params_def,
            params_str="[[mot1] [mot2 3]]",
            params=[[["mot1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt7d2_params_def,
            params_str="[[mot1 1] [mot2 3]]",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt7d2_params_def,
            params_str="mot1 1 mot2 3",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt7d2_params_def,
            params_str="[[] [mot2 3] []]",
            params=[[[], ["mot2", "3"], []]])
@insertTest(helper_name="parse", params_def=pt8_params_def,
            params_str="[[mot1 1] [mot2 3]]",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt8_params_def,
            params_str="mot1 1 mot2 3",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt9_params_def,
            params_str="[[mot1 1] [mot2 3]]",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt9_params_def,
            params_str="mot1 1 mot2 3",
            params=[[["mot1", "1"], ["mot2", "3"]]])
@insertTest(helper_name="parse", params_def=pt10_params_def,
            params_str="[1 3] mot1", params=[["1", "3"], "mot1"])
@insertTest(helper_name="parse", params_def=pt10_params_def,
            params_str="1 mot1", params=[["1"], "mot1"])
@insertTest(helper_name="parse", params_def=pt11_params_def,
            params_str="ct1 [1 3] mot1", params=["ct1", ["1", "3"], "mot1"])
@insertTest(helper_name="parse", params_def=pt12_params_def,
            params_str="[1 3 4] [mot1 mot2]",
            params=[["1", "3", "4"], ["mot1", "mot2"]])
@insertTest(helper_name="parse", params_def=pt13_params_def,
            params_str="[[mot1 mot2] [mot3 mot4]]",
            params=[[["mot1", "mot2"], ["mot3", "mot4"]]])
@insertTest(helper_name="parse", params_def=pt14_params_def,
            params_str="[[[mot1 mot2] 3] [[mot3] 5]]",
            params=[[[["mot1", "mot2"], "3"], [["mot3"], "5"]]])
@insertTest(helper_name="parse", params_def=pt14d_params_def,
            params_str="[[[mot1 mot2] 3] [[mot3] []]]",
            params=[[[["mot1", "mot2"], "3"], [["mot3"], []]]])
@insertTest(helper_name="parse", params_def=pt14d_params_def,
            params_str="[[[mot1 []] 3] [[mot3] []]]",
            params=[[[["mot1", []], "3"], [["mot3"], []]]])
@insertTest(helper_name="parse", params_def=pt14d_params_def,
            params_str="[[[[]] 3] [[mot3] []]]",
            params=[[[[[]], "3"], [["mot3"], []]]])
# extra tests for complex parameter values
@insertTest(helper_name="parse", params_def=extra1_params_def,
            params_str="ScanFile ['file.nxs' 'file.dat']",
            params=["ScanFile", ["file.nxs", "file.dat"]])
@insertTest(helper_name="parse", params_def=extra2_params_def,
            params_str="2 3 ['Hello world!' 'How are you?']",
            params=["2", "3", ["Hello world!", "How are you?"]])
@insertTest(helper_name="parse", params_def=extra3_params_def,
            params_str="ScanFile file.dat",
            params=["ScanFile", "file.dat"])
@insertTest(helper_name="parse", params_def=extra4_params_def,
            params_str="'2 3'", params=["2 3"])
@insertTest(helper_name="parse", params_def=extra5_params_def,
            params_str="[[mot01 3][mot02 5]] ct01 999",
            params=[[["mot01", "3"], ["mot02", "5"]], "ct01", "999"])
@insertTest(helper_name="parse", params_def=extra6_params_def,
            params_str="[[2 3][4 5]]",
            params=[[["2", "3"], ["4", "5"]]])
@insertTest(helper_name="parse", params_def=extra7_params_def,
            params_str="1 [2 3]",
            params=["1", ["2", "3"]])
@insertTest(helper_name="parse", params_def=extra8_params_def,
            params_str="2 3", params=["2", "3"])
class ParamParserTestCase(unittest.TestCase):
    """Unit tests for ParamParser class. Mainly based on macro examples for
    parameters definition.
    """
    def parse(self, params_def, params_str, params):
        """Helper method to test parameters parsing. To be used with
        insertTest decorator.
        """
        p = ParamParser(params_def)
        result = p.parse(params_str)
        msg = "Parsing failed (result: %r; expected: %r)" % \
              (result, params)
        self.assertListEqual(result, params, msg)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import PyTango
import unittest
from taurus.test import insertTest
from sardana.tango.pool.test import BasePoolTestCase
from sardana.tango.core.util import get_free_alias

info1 = ('Motor', 'DummyMotorController', 'DummyMotorController')
info2 = ('TriggerGate', 'DummyTriggerGateController',
         'DummyTriggerGateController')


@insertTest(helper_name='check_elems_presistence',
            test_method_doc='Test persistence of dummy Motor elements',
            info=info1)
@insertTest(helper_name='check_elems_presistence',
            test_method_doc='Test persistence of dummy TriggerGate elements',
            info=info2)
class PersistenceTestCase(BasePoolTestCase, unittest.TestCase):
    """ Test the persistence of the Sardana Tango elements.
    """

    def check_elems_presistence(self, info):
        """Helper method to test the elements persistence. The actions are:
            - creation of controller and element
            - restart Pool
            - check if element persist

        :param info: information about controller (type, library, class)
        :type info: tuple<str>
        """
        # Create controller
        self.do_element_cleanup = False
        sar_type, lib, cls = info
        base_name = "ctrl_persistent_" + cls
        self.ctrl_name = get_free_alias(PyTango.Database(), base_name)
        self.pool.CreateController([sar_type, lib, cls, self.ctrl_name])

        # Create element
        base_name = "elem_persistent_" + cls
        self.elem_name = get_free_alias(PyTango.Database(), base_name)
        axis = 1
        self.pool.CreateElement([sar_type, self.ctrl_name, str(axis),
                                 self.elem_name])
        # Restart Pool
        self._starter.stopDs(hard_kill=True)
        self._starter.startDs(wait_seconds=20)
        # Check if the element exists
        try:
            obj = PyTango.DeviceProxy(self.elem_name)
            # the element is persistent, cleanup is necessary
            self.do_element_cleanup = True
        except:
            obj = None
        msg = 'The element "%s" does not exist after restarting the Pool' %\
            (self.elem_name)
        self.assertIsNotNone(obj, msg)

    def tearDown(self):
        """Remove the elements and the controllers
        """
        cleanup_success = True
        if self.do_element_cleanup:
            try:
                self.pool.DeleteElement(self.elem_name)
            except:
                cleanup_success = False
        try:
            self.pool.DeleteElement(self.ctrl_name)
        except:
            cleanup_success = False
        BasePoolTestCase.tearDown(self)
        if not cleanup_success:
            raise Exception("Cleanup failed. Database may be left dirty.")



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Taurus Qt extensions for Sardana devices.

Objects obtained with :func:`taurus.Device` expose standard interfaces
e.g., allow to interact with their attributes, check their state, etc.
This module defines classes for enriched interaction with Sardana devices
(also for other elements not exported as devices), e.g. synchronous move
of a sardana motor with
:meth:`~sardana.taurus.core.tango.sardana.pool.Motor.move`
method instead of writing motor's position attribute and then waiting for its
state change. The difference between these classes with respect to the ones
from the :mod:`sardana.taurus.core.tango.sardana` module is the Qt friendly
interface e.g. the Sardana events are translated to Qt signals.

To obtain these enriched objects with :func:`taurus.Device` you need to first
register the extension classes with
the :obj:`~sardana.taurus.qt.qtcore.tango.sardana.registerExtensions` function.

The registration needs to be done before the first access to the given
:func:`taurus.Device`.

.. note:: If you are using
  :class:`~taurus.qt.qtgui.application.TaurusApplication`
  then the registration is done behind the scene at the moment of
  :class:`~taurus.qt.qtgui.application.TaurusApplication` construction.
"""

__docformat__ = 'restructuredtext'


def registerExtensions():
    from . import pool
    from . import macroserver

    pool.registerExtensions()
    macroserver.registerExtensions()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the definition of the macroserver parameters for
macros"""

__all__ = ["WrongParam", "MissingParam", "SupernumeraryParam",
           "UnknownParamObj", "WrongParamType", "MissingRepeat",
           "SupernumeraryRepeat", "TypeNames", "Type", "ParamType",
           "ElementParamType", "ElementParamInterface",
           "AttrParamType", "AbstractParamTypes", "ParamDecoder"]

__docformat__ = 'restructuredtext'

from copy import deepcopy
from lxml import etree
from taurus.core.util.containers import CaselessDict

from sardana import ElementType, INTERFACES_EXPANDED
from sardana.sardanautils import is_non_str_seq
from sardana.macroserver.msbase import MSBaseObject
from sardana.macroserver.msexception import MacroServerException, \
    UnknownMacro, UnknownMacroLibrary


class OptionalParamClass(dict):
    def __init__(self, obj):
        super(OptionalParamClass, self).__init__(obj)
        attributes = dir(self)

        for attr in attributes:
            # items is necessary fo python 3.5 implementation of json
            if attr in ['__setattr__', 'raise_error', '__class__',
                        '__dict__', '__weakref__', 'items']:
                continue
            self.__setattr__(attr, self.raise_error)
        self.__setattr__ = self.raise_error

    def __repr__(self):
        return 'Optional'

    def raise_error(*args, **kwargs):
        raise RuntimeError('can not be accessed')


Optional = OptionalParamClass({'___optional_parameter__': True})


class WrongParam(MacroServerException):

    def __init__(self, *args):
        MacroServerException.__init__(self, *args)
        self.type = 'Wrong parameter'


class MissingParam(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Missing parameter'


class SupernumeraryParam(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Supernumerary parameter'


class UnknownParamObj(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Unknown parameter'


class WrongParamType(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Unknown parameter type'


class MissingRepeat(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Missing repeat'


class SupernumeraryRepeat(WrongParam):

    def __init__(self, *args):
        WrongParam.__init__(self, *args)
        self.type = 'Supernumerary repeat'


class TypeNames:
    """Class that holds the list of registered macro parameter types"""

    def __init__(self):
        self._type_names = {}
        self._pending_type_names = {}

    def addType(self, name):
        """Register a new macro parameter type"""
        setattr(self, name, name)
        self._type_names[name] = name
        if name in self._pending_type_names:
            del self._pending_type_names[name]

    def removeType(self, name):
        """remove a macro parameter type"""
        delattr(self, name)
        try:
            del self._type_names[name]
        except ValueError:
            pass

    def __str__(self):
        return str(list(self._type_names.keys()))

#    def __getattr__(self, name):
#        if name not in self._pending_type_names:
#            self._pending_type_names[name] = name
#        return self._pending_type_names[name]


# This instance of TypeNames is intended to provide access to types to the
# Macros in a "Type.Motor" fashion
Type = TypeNames()


class ParamType(MSBaseObject):

    All = 'All'

    # Capabilities
    ItemList = 'ItemList'
    ItemListEvents = 'ItemListEvents'

    capabilities = []

    type_class = str

    def __init__(self, macro_server, name):
        MSBaseObject.__init__(self, name=name, full_name=name,
                              macro_server=macro_server,
                              elem_type=ElementType.ParameterType)

    def getName(self):
        return self.name

    def getObj(self, str_repr):
        return self.type_class(str_repr)

    @classmethod
    def hasCapability(cls, cap):
        return cap in cls.capabilities

    def serialize(self, *args, **kwargs):
        kwargs = MSBaseObject.serialize(self, *args, **kwargs)
        kwargs['composed'] = False
        return kwargs


class ElementParamType(ParamType):

    capabilities = ParamType.ItemList, ParamType.ItemListEvents

    def __init__(self, macro_server, name):
        ParamType.__init__(self, macro_server, name)

    def accepts(self, elem):
        return elem.getType() == self._name

    def getObj(self, name, pool=ParamType.All, cache=False):
        macro_server = self.macro_server
        if pool == ParamType.All:
            pools = macro_server.get_pools()
        else:
            pools = macro_server.get_pool(pool),
        for pool in pools:
            elem_info = pool.getObj(name, elem_type=self._name)
            if elem_info is not None and self.accepts(elem_info):
                return elem_info
        # not a pool object, maybe it is a macro server object (perhaps a macro
        # code or a macro library
        try:
            return macro_server.get_macro(name)
        except UnknownMacro:
            pass

        try:
            return macro_server.get_macro_lib(name)
        except UnknownMacroLibrary:
            pass
        # neither pool nor macroserver contains any element with this name
        raise UnknownParamObj('%s with name %s does not exist' %
                              (self._name, name))

    def getObjDict(self, pool=ParamType.All, cache=False):
        macro_server = self.macro_server
        objs = CaselessDict()
        if pool == ParamType.All:
            pools = macro_server.get_pools()
        else:
            pools = macro_server.get_pool(pool),
        for pool in pools:
            for elem_info in pool.getElements():
                if self.accepts(elem_info):
                    objs[elem_info.name] = elem_info
        macros = macro_server.get_macros()
        for macro_lib_name, macro_lib in list(macros.items()):
            if self.accepts(macro_lib):
                objs[macro_lib_name] = macro_lib
        for macro_name, macro in list(macro_server.get_macros().items()):
            if self.accepts(macro):
                objs[macro_name] = macro

        return objs

    def getObjListStr(self, pool=ParamType.All, cache=False):
        obj_dict = self.getObjDict(pool=pool, cache=cache)
        return list(obj_dict.keys())

    def getObjList(self, pool=ParamType.All, cache=False):
        obj_dict = self.getObjDict(pool=pool, cache=cache)
        return list(obj_dict.values())

    def serialize(self, *args, **kwargs):
        kwargs = ParamType.serialize(self, *args, **kwargs)
        kwargs['composed'] = True
        return kwargs


class ElementParamInterface(ElementParamType):

    def __init__(self, macro_server, name):
        ElementParamType.__init__(self, macro_server, name)
        bases, doc = INTERFACES_EXPANDED.get(name)
        self._interfaces = bases

    def accepts(self, elem):
        elem_type = elem.getType()
        elem_interfaces = INTERFACES_EXPANDED.get(elem_type)[0]
        if elem_interfaces is None:
            return ElementParamType.accepts(self, elem)
        return self._name in elem_interfaces

    def getObj(self, name, pool=ParamType.All, cache=False):
        macro_server = self.macro_server
        if pool == ParamType.All:
            pools = macro_server.get_pools()
        else:
            pools = macro_server.get_pool(pool),
        for pool in pools:
            elem_info = pool.getElementWithInterface(name, self._name)
            if elem_info is not None and self.accepts(elem_info):
                return elem_info
        # not a pool object, maybe it is a macro server object (perhaps a macro
        # class or a macro library
        try:
            return macro_server.get_macro(name)
        except UnknownMacro:
            pass

        try:
            return macro_server.get_macro_lib(name)
        except UnknownMacroLibrary:
            pass
        # neither pool nor macroserver contains any element with this name
        raise UnknownParamObj('%s with name %s does not exist' %
                              (self._name, name))

    def getObjDict(self, pool=ParamType.All, cache=False):
        macro_server = self.macro_server
        objs = CaselessDict()
        if macro_server.is_macroserver_interface(self._name):
            return macro_server.get_elements_with_interface(self._name)

        if pool == ParamType.All:
            pools = macro_server.get_pools()
        else:
            pools = macro_server.get_pool(pool),
        for pool in pools:
            elements = pool.getElementsWithInterface(self._name)
            for elem_info in list(elements.values()):
                if self.accepts(elem_info):
                    objs[elem_info.name] = elem_info
        return objs

    def getObjListStr(self, pool=ParamType.All, cache=False):
        obj_dict = self.getObjDict(pool=pool, cache=cache)
        return list(obj_dict.keys())

    def getObjList(self, pool=ParamType.All, cache=False):
        obj_dict = self.getObjDict(pool=pool, cache=cache)
        return list(obj_dict.values())


class AttrParamType(ParamType):
    pass


AbstractParamTypes = ParamType, ElementParamType, ElementParamInterface, AttrParamType


class ParamDecoder:

    def __init__(self, type_manager, params_def, raw_params):
        """Create ParamDecorder object and decode macro parameters

        :param type_manager: (sardana.macroserver.mstypemanager.TypeManager)
            type manager object
        :param params_def: list<list> macro parameter definition
        :param raw_params: (lxml.etree._Element or list) xml element
            representing macro with subelements representing parameters or list
            with parameter values
        """
        self.type_manager = type_manager
        self.params_def = params_def
        self.raw_params = raw_params
        self.params = None
        self.decode()

    def decode(self):
        """Decode raw representation of parameters to parameters as passed
        to the prepare or run methods.
        """
        # make a copy since in case of XML it could be necessary to modify
        # the raw_params - filter out elements different than params
        raw_params = deepcopy(self.raw_params)
        params_def = self.params_def
        # ignore other tags than "param" and "paramRepeat"
        # e.g. sequencer may create tags like "hookPlace"
        if isinstance(raw_params, etree._Element):
            for raw_param in raw_params:
                if not raw_param.tag in ("param", "paramrepeat"):
                    raw_params.remove(raw_param)

        params = []
        # check if too many parameters were passed
        len_params_def = len(params_def)
        if len(raw_params) > len_params_def:
            msg = ("%r are supernumerary with respect to definition" %
                   raw_params[len_params_def:])
            raise SupernumeraryParam(msg)
        # iterate over definition since missing values may just mean using
        # the default values
        for i, param_def in enumerate(params_def):
            try:
                raw_param = raw_params[i]
            except IndexError:
                raw_param = None
            obj = self.decodeNormal(raw_param, param_def)
            params.append(obj)
        self.params = params
        return self.params

    def decodeNormal(self, raw_param, param_def):
        """Decode and validate parameter

        :param raw_param: (lxml.etree._Element or list) xml element
            representing parameter
        :param param_def: (dict) parameter definition

        :return: (list): list with decoded parameter repetitions
        """
        param_type = param_def["type"]
        name = param_def["name"]
        optional_param = False
        if isinstance(param_type, list):
            param = self.decodeRepeat(raw_param, param_def)
        else:
            type_manager = self.type_manager
            param_type = type_manager.getTypeObj(param_type)
            try:
                if isinstance(raw_param, etree._Element):
                    value = raw_param.get("value")
                else:
                    value = raw_param
                # None or [] indicates default value
                if value is None or (isinstance(value, list) and
                                     len(value) == 0):
                    value = param_def['default_value']
                if value is None:
                    raise MissingParam("'%s' not specified" % name)
                elif value is Optional:
                    param = None
                    optional_param = True
                else:
                    # cast to sting to fulfill with ParamType API
                    param = param_type.getObj(str(value))

            except ValueError as e:
                raise WrongParamType(str(e)) from e
            except UnknownParamObj as e:
                raise WrongParam(str(e)) from e
            if param is None and not optional_param:
                msg = 'Could not create %s parameter "%s" for "%s"' % \
                      (param_type.getName(), name, raw_param)
                raise WrongParam(msg)
        return param

    def decodeRepeat(self, raw_param_repeat, param_repeat_def):
        """Decode and validate repeat parameter

        :param raw_param_repeat: (lxml.etree._Element or list) xml element
            representing param repeat with subelements representing repetitions
            or list representing repetitions
        :param param_repeat_def: (dict) repeat parameter definition

        :return: (list): list with decoded parameter repetitions
        """
        name = param_repeat_def['name']
        param_type = param_repeat_def['type']
        min_rep = param_repeat_def['min']
        max_rep = param_repeat_def['max']
        param_repeat = []
        if raw_param_repeat is None:
            raw_param_repeat = param_repeat_def['default_value']
        if raw_param_repeat is None:
            raw_param_repeat = []
        len_rep = len(raw_param_repeat)
        if min_rep and len_rep < min_rep:
            msg = 'Found %d repetitions of param %s, min is %d' % \
                  (len_rep, name, min_rep)
            raise MissingRepeat(msg)
        if max_rep and len_rep > max_rep:
            msg = 'Found %d repetitions of param %s, max is %d' % \
                  (len_rep, name, max_rep)
            raise SupernumeraryRepeat(msg)
        # repeat params with only one member and only one repetition value are
        # allowed - encapsulate it in list and try to decode anyway;
        # for the moment this only works for non XML decoding but could be
        # extended in the future to support XML as well
        if not is_non_str_seq(raw_param_repeat)\
                and not isinstance(raw_param_repeat, etree._Element):
            raw_param_repeat = [raw_param_repeat]
        for raw_repeat in raw_param_repeat:
            if len(param_type) > 1:
                repeat = []
                for i, member_type in enumerate(param_type):
                    try:
                        member_raw = raw_repeat[i]
                    except IndexError:
                        member_raw = None
                    member = self.decodeNormal(member_raw, member_type)
                    repeat.append(member)
            else:
                # if the repeat parameter is composed of just one member
                # do not encapsulate it in list and pass directly the item
                if isinstance(raw_repeat, etree._Element):
                    raw_repeat = raw_repeat[0]
                # check if one tries to decode repeat parameter of just one
                # member encapsulated in a list, empty lists are still allowed
                # to indicate default value
                elif isinstance(raw_repeat, list) and len(raw_repeat) > 0:
                    msg = 'Repetitions of just one member must not be lists'
                    raise WrongParam(msg)
                repeat = self.decodeNormal(raw_repeat, param_type[0])
            param_repeat.append(repeat)
        return param_repeat

    def getParamList(self):
        return self.params

    def __iter__(self):
        return iter(self.params)

    def __getattr__(self, name):
        return getattr(self.params, name)


class FlatParamDecoder:
    """Parameter decoder useful for macros with only one repeat parameter
    located at the very last place. It requires that the raw parameters are
    passed as a flat list of strings.
    """

    def __init__(self, type_manager, params_def, raw_params):
        self.type_manager = type_manager
        self.params_def = params_def
        self.raw_params = raw_params
        self.params = None
        if not self.isPossible(params_def):
            msg = ("%s parameter definition is not compatible with"
                   " FlatParamDecoder" % params_def)
            raise AttributeError(msg)
        self.decode()

    @staticmethod
    def isPossible(params_def):
        for param_def in params_def:
            param_type = param_def["type"]
            if isinstance(param_type, list):
                if param_def != params_def[-1]:
                    # repeat parameter is not the last one
                    # it won't be possible to decode it
                    return False
                else:
                    for sub_param_def in param_type:
                        if isinstance(sub_param_def, list):
                            # nested repeat parameter
                            # it won't be possible to decode it
                            return False
        return True

    def decode(self):
        params_def = self.params_def
        raw_params = self.raw_params
        _, self.params = self.decodeNormal(raw_params, params_def)
        return self.params

    def decodeNormal(self, raw_params, params_def):
        str_len = len(raw_params)
        obj_list = []
        str_idx = 0
        for i, par_def in enumerate(params_def):
            name = par_def['name']
            type_class = par_def['type']
            def_val = par_def['default_value']
            if str_idx == str_len:
                if def_val is None:
                    if not isinstance(type_class, list):
                        raise MissingParam("'%s' not specified" % name)
                    elif isinstance(type_class, list):
                        min_rep = par_def['min']
                        if min_rep > 0:
                            msg = "'%s' demands at least %d values" %\
                                  (name, min_rep)
                            raise WrongParam(msg)
                if not def_val is None:
                    new_obj = def_val
            else:
                if isinstance(type_class, list):
                    data = self.decodeRepeat(raw_params[str_idx:], par_def)
                    dec_token, new_obj = data
                else:
                    type_manager = self.type_manager
                    type_name = type_class
                    type_class = type_manager.getTypeClass(type_name)
                    par_type = type_manager.getTypeObj(type_name)
                    par_str = raw_params[str_idx]
                    try:
                        val = par_type.getObj(par_str)
                    except ValueError as e:
                        raise WrongParamType(str(e)) from e
                    except UnknownParamObj as e:
                        raise WrongParam(str(e)) from e
                    if val is None:
                        msg = 'Could not create %s parameter "%s" for "%s"' % \
                              (par_type.getName(), name, par_str)
                        raise WrongParam(msg)
                    dec_token = 1
                    new_obj = val
                str_idx += dec_token
            obj_list.append(new_obj)
        return str_idx, obj_list

    def decodeRepeat(self, raw_params, par_def):
        name = par_def['name']
        param_def = par_def['type']
        min_rep = par_def['min']
        max_rep = par_def['max']

        dec_token = 0
        obj_list = []
        rep_nr = 0
        while dec_token < len(raw_params):
            if max_rep is not None and rep_nr == max_rep:
                break
            new_token, new_obj_list = self.decodeNormal(raw_params[dec_token:],
                                                        param_def)
            dec_token += new_token
            if len(new_obj_list) == 1:
                new_obj_list = new_obj_list[0]
            obj_list.append(new_obj_list)
            rep_nr += 1
        if rep_nr < min_rep:
            msg = 'Found %d repetitions of param %s, min is %d' % \
                  (rep_nr, name, min_rep)
            raise MissingRepeat(msg)
        return dec_token, obj_list

    def getParamList(self):
        return self.params

    def __iter__(self):
        return iter(self.params)

    def __getattr__(self, name):
        return getattr(self.params, name)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""The sardana tango measurement group module"""

__all__ = ["MeasurementGroup", "MeasurementGroupClass"]

__docformat__ = 'restructuredtext'

import sys
import time

from PyTango import Except, DevVoid, DevLong, DevDouble, DevString, \
    DispLevel, DevState, AttrQuality, READ, READ_WRITE, SCALAR, Util

from taurus.core.util.codecs import CodecFactory
from taurus.core.util.log import DebugIt

from sardana import State, SardanaServer
from sardana.sardanaattribute import SardanaAttribute
from sardana.pool import AcqMode
from sardana.pool.pooldefs import SynchDomain, SynchParam
from sardana.tango.core.util import exception_str
from sardana.tango.pool.PoolDevice import PoolGroupDevice, PoolGroupDeviceClass


class MeasurementGroup(PoolGroupDevice):

    def __init__(self, dclass, name):
        PoolGroupDevice.__init__(self, dclass, name)

    def init(self, name):
        PoolGroupDevice.init(self, name)

    def get_measurement_group(self):
        return self.element

    def set_measurement_group(self, measurement_group):
        self.element = measurement_group

    measurement_group = property(get_measurement_group, set_measurement_group)

    @DebugIt()
    def delete_device(self):
        PoolGroupDevice.delete_device(self)
        mg = self.measurement_group
        if mg is not None:
            mg.remove_listener(self.on_measurement_group_changed)

    @DebugIt()
    def init_device(self):
        PoolGroupDevice.init_device(self)
        # state and status are already set by the super class
        detect_evts = "moveable", "synchdescription", \
                      "softwaresynchronizerinitialdomain"
        # TODO: nbstarts could be moved to detect events with
        # abs_change criteria of 1, but be careful with
        # tango-controls/pytango#302
        non_detect_evts = "configuration", "integrationtime", "monitorcount", \
                          "acquisitionmode", "elementlist", "latencytime", \
                          "nbstarts"
        self.set_change_events(detect_evts, non_detect_evts)

        self.Elements = list(self.Elements)
        for i in range(len(self.Elements)):
            try:
                self.Elements[i] = int(self.Elements[i])
            except:
                pass
        mg = self.measurement_group
        if mg is None:
            full_name = self.get_full_name()
            name = self.alias or full_name
            self.measurement_group = mg = \
                self.pool.create_measurement_group(name=name,
                                                   full_name=full_name,
                                                   id=self.Id,
                                                   user_elements=self.Elements)
        mg.add_listener(self.on_measurement_group_changed)

        # force a state read to initialize the state attribute
        # state = self.measurement_group.state
        self.set_state(DevState.ON)

    def on_measurement_group_changed(self, event_source, event_type,
                                     event_value):
        try:
            self._on_measurement_group_changed(
                event_source, event_type, event_value)
        except:
            msg = 'Error occured "on_measurement_group_changed(%s.%s): %s"'
            exc_info = sys.exc_info()
            self.error(msg, self.measurement_group.name, event_type.name,
                       exception_str(*exc_info[:2]))
            self.debug("Details", exc_info=exc_info)

    def _on_measurement_group_changed(self, event_source, event_type,
                                      event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        timestamp = time.time()
        name = event_type.name
        name = name.replace('_', '')
        multi_attr = self.get_device_attr()
        attr = multi_attr.get_attr_by_name(name)
        quality = AttrQuality.ATTR_VALID
        priority = event_type.priority
        error = None

        if name == "state":
            event_value = self.calculate_tango_state(event_value)
        elif name == "status":
            event_value = self.calculate_tango_status(event_value)
        elif name == "acquisitionmode":
            event_value = AcqMode.whatis(event_value)
        elif name == "configuration":
            cfg = self.measurement_group.get_user_configuration()
            codec = CodecFactory().getCodec('json')
            _, event_value = codec.encode(('', cfg))
        elif name == "synchdescription":
            codec = CodecFactory().getCodec('json')
            _, event_value = codec.encode(('', event_value))
        elif name == "moveable" and event_value is None:
            event_value = 'None'
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                timestamp = event_value.timestamp
                event_value = event_value.value

        self.set_attribute(attr, value=event_value, timestamp=timestamp,
                           quality=quality, priority=priority, error=error,
                           synch=False)

    def _synch_description_str2enum(self, _synch_description_str):
        """Translates synchronization description data structure so it uses
        SynchParam and SynchDomain enums as keys instead of strings.

        .. todo:: At some point remove the backwards compatibility
          for memorized values created with Python 2. In Python 2 IntEnum was
          serialized to "<class>.<attr>" e.g. "SynchDomain.Time" and we were
          using a class method `fromStr` to interpret the enumeration objects.
        """
        synch_description = []
        for group_str in _synch_description_str:
            group = {}
            for param_str, conf_str in group_str.items():
                try:
                    param = SynchParam(int(param_str))
                except ValueError:
                    param = SynchParam.fromStr(param_str)
                if isinstance(conf_str, dict):
                    conf = {}
                    for domain_str, value in conf_str.items():
                        try:
                            domain = SynchDomain(int(domain_str))
                        except ValueError:
                            domain = SynchDomain.fromStr(domain_str)
                        conf[domain] = value
                else:
                    conf = conf_str
                group[param] = conf
            synch_description.append(group)
        return synch_description

    def always_executed_hook(self):
        pass
        # state = to_tango_state(self.motor_group.get_state(cache=False))

    def read_attr_hardware(self, data):
        pass

    def read_IntegrationTime(self, attr):
        it = self.measurement_group.integration_time
        if it is None:
            it = float('nan')
        attr.set_value(it)

    def write_IntegrationTime(self, attr):
        self.measurement_group.integration_time = attr.get_write_value()

    def read_MonitorCount(self, attr):
        it = self.measurement_group.monitor_count
        if it is None:
            it = 0
        attr.set_value(it)

    def write_MonitorCount(self, attr):
        self.measurement_group.monitor_count = attr.get_write_value()

    def read_AcquisitionMode(self, attr):
        acq_mode = self.measurement_group.acquisition_mode
        acq_mode_str = AcqMode.whatis(acq_mode)
        attr.set_value(acq_mode_str)

    def write_AcquisitionMode(self, attr):
        acq_mode_str = attr.get_write_value()
        try:
            acq_mode = AcqMode.lookup[acq_mode_str]
        except KeyError:
            raise Exception("Invalid acquisition mode. Must be one of " +
                            ", ".join(list(AcqMode.keys())))
        self.measurement_group.acquisition_mode = acq_mode

    def read_Configuration(self, attr):
        cfg = self.measurement_group.get_user_configuration()
        codec = CodecFactory().getCodec('json')
        data = codec.encode(('', cfg))
        attr.set_value(data[1])

    def write_Configuration(self, attr):
        data = attr.get_write_value()
        cfg = CodecFactory().decode(('json', data))
        util = Util.instance()
        if util.is_svr_starting():
            self.measurement_group._config._value_ref_compat = True
        else:
            self.measurement_group._config._value_ref_compat = False
        self.measurement_group.set_configuration_from_user(cfg)

    def read_NbStarts(self, attr):
        nb_starts = self.measurement_group.nb_starts
        if nb_starts is None:
            nb_starts = int('nan')
        attr.set_value(nb_starts)

    def write_NbStarts(self, attr):
        self.measurement_group.nb_starts = attr.get_write_value()

    def read_Moveable(self, attr):
        moveable = self.measurement_group.moveable
        if moveable is None:
            moveable = 'None'
        attr.set_value(moveable)

    def write_Moveable(self, attr):
        moveable = attr.get_write_value()
        if moveable == 'None':
            moveable = None
        self.measurement_group.moveable = moveable

    def read_SynchDescription(self, attr):
        synch_description = self.measurement_group.synch_description
        codec = CodecFactory().getCodec('json')
        data = codec.encode(('', synch_description))
        attr.set_value(data[1])

    def write_SynchDescription(self, attr):
        data = attr.get_write_value()
        synch_description = CodecFactory().decode(('json', data))
        # translate dictionary keys
        synch_description = \
            self._synch_description_str2enum(synch_description)
        self.measurement_group.synch_description = synch_description

    def read_LatencyTime(self, attr):
        latency_time = self.measurement_group.latency_time
        attr.set_value(latency_time)

    def read_SoftwareSynchronizerInitialDomain(self, attr):
        domain = self.measurement_group.sw_synch_initial_domain
        d = SynchDomain(domain).name
        attr.set_value(d)

    def write_SoftwareSynchronizerInitialDomain(self, attr):
        data = attr.get_write_value()
        try:
            domain = SynchDomain[data]
        except KeyError:
            raise Exception("Invalid domain (can be either Position or Time)")
        self.measurement_group.sw_synch_initial_domain = domain

    def Prepare(self):
        self.measurement_group.prepare()

    def Start(self):
        try:
            self.wait_for_operation()
        except:
            raise Exception("Cannot acquire: already involved in an operation")
        self.measurement_group.start_acquisition()

    def Stop(self):
        self.measurement_group.stop()


class MeasurementGroupClass(PoolGroupDeviceClass):

    #    Class Properties
    class_property_list = {
    }

    #    Device Properties
    device_property_list = {
    }
    device_property_list.update(PoolGroupDeviceClass.device_property_list)

    #    Command definitions
    cmd_list = {
        'Prepare': [[DevVoid, ""], [DevVoid, ""]],
        'Start': [[DevVoid, ""], [DevVoid, ""]]
    }
    cmd_list.update(PoolGroupDeviceClass.cmd_list)

    #    Attribute definitions
    attr_list = {
        'IntegrationTime': [[DevDouble, SCALAR, READ_WRITE],
                            {'Memorized': "true",
                             'Display level': DispLevel.OPERATOR}],
        'MonitorCount': [[DevLong, SCALAR, READ_WRITE],
                         {'Memorized': "true",
                          'Display level': DispLevel.OPERATOR}],
        'AcquisitionMode': [[DevString, SCALAR, READ_WRITE],
                            {'Memorized': "true",
                             'Display level': DispLevel.OPERATOR}],
        'Configuration': [[DevString, SCALAR, READ_WRITE],
                          {'Memorized': "true",
                           'Display level': DispLevel.EXPERT}],
        'NbStarts': [[DevLong, SCALAR, READ_WRITE],
                     {'Memorized': "true",
                      'Display level': DispLevel.OPERATOR}],
        'Moveable': [[DevString, SCALAR, READ_WRITE],
                     {'Memorized': "true",
                      'Display level': DispLevel.EXPERT}],
        # TODO: Does it have sense to memorize SynchDescription?
        'SynchDescription': [[DevString, SCALAR, READ_WRITE],
                             {'Memorized': "true",
                              'Display level': DispLevel.EXPERT}],
        'LatencyTime': [[DevDouble, SCALAR, READ],
                        {'Display level': DispLevel.EXPERT}],
        'SoftwareSynchronizerInitialDomain': [[DevString, SCALAR, READ_WRITE],
                                              {'Memorized': "true",
                                               'Display level':
                                               DispLevel.OPERATOR}],

    }
    attr_list.update(PoolGroupDeviceClass.attr_list)

    def _get_class_properties(self):
        ret = PoolGroupDeviceClass._get_class_properties(self)
        ret['Description'] = "Measurement group device class"
        ret['InheritedFrom'].insert(0, 'PoolGroupDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Tests for continuous scans (ct-like)"""
import time
import PyTango
import unittest
from sardana.macroserver.macros.test import (RunStopMacroTestCase, testRun,
                                             testStop)
from sardana.pool import AcqSynchType
# TODO: MeasSarTestTestCase is a util, could be moved to base_sartest
# or another utils module.
from sardana.tango.pool.test.test_measurementgroup import MeasSarTestTestCase
from sardana.tango.macroserver.test import BaseMacroServerTestCase


class UtilsForTests():

    def parsingOutputPoints(self, log_output):
        """A helper method to know if points are ordered based on log_output.
        """
        first_data_line = 1
        scan_index = 0
        list_points = []
        for line, in log_output[first_data_line:]:
            # Get a list of elements without white spaces between them
            columns = line.split()

            # Cast index of scan to int (the first element of the list)
            columns[scan_index] = int(columns[scan_index])
            list_points.append(columns[scan_index])
        nb_points = len(list_points)

        ordered_points = 0
        for i in range(len(list_points) - 1):
            if list_points[i + 1] >= list_points[i]:
                ordered_points = 1
            else:
                ordered_points = 0
                break

        return (nb_points, ordered_points)

    def orderPointsData(self, data):
        """A helper method to know if points are ordered based on getData.
        """
        obtained_nb_points_data = len(list(data.keys()))
        ordered_points_data = 0
        for i in range(obtained_nb_points_data - 1):
            if int(list(data.keys())[i + 1]) >= int(list(data.keys())[i]):
                ordered_points_data = 1
            else:
                ordered_points_data = 0
                break
        return ordered_points_data


class ScanctTest(MeasSarTestTestCase, BaseMacroServerTestCase,
                 RunStopMacroTestCase):
    """Base class for the continuous scans (ct-like) tests. Implements
    methods for preparation of the elements and validation of the results.
    """

    utils = UtilsForTests()

    def setUp(self):
        MeasSarTestTestCase.setUp(self)
        properties = {'PoolNames': self.pool_name}
        BaseMacroServerTestCase.setUp(self, properties)
        RunStopMacroTestCase.setUp(self)

    def configure_motors(self, motor_names):
        # TODO: workaround for bug with velocity<base_rate: Sdn#38
        for name in motor_names:
            mot = PyTango.DeviceProxy(name)
            mot.write_attribute('acceleration', 0.1)
            mot.write_attribute('base_rate', 0)
            mot.write_attribute('deceleration', 0.1)

    def configure_mntgrp(self, meas_config):
        # creating MEAS
        self.create_meas(meas_config)
        # Set ActiveMntGrp
        self.macro_executor.run(macro_name='senv',
                                macro_params=['ActiveMntGrp', '_test_mg_1'],
                                sync=True, timeout=1.)

    def check_using_output(self, expected_nb_points):
        # Test data from log_output
        log_output = self.macro_executor.getLog('output')
        (aa, bb) = self.utils.parsingOutputPoints(log_output)
        # ordered_points: (int) obtained number of points.
        obtained_nb_points = aa
        # ordered_points: booleand which indicates if points are ordered.
        ordered_points = bb

        msg = ("The ascanct execution did not return any scan point.\n"
               "Checked using macro output")
        self.assertNotEqual(obtained_nb_points, 0, msg)

        msg = ("The ascanct execution did not return the expected number of "
               "points.\nExpected " + str(expected_nb_points) + " points."
               "\nObtained " + str(obtained_nb_points) + " points."
               "Checked using macro output")
        self.assertEqual(obtained_nb_points, expected_nb_points, msg)

        msg = "Scan points are NOT in good order.\nChecked using macro output"
        self.assertTrue(ordered_points, msg)

    def check_using_data(self, expected_nb_points):
        # Test data from macro (macro_executor.getData())
        data = self.macro_executor.getData()
        order_points_data = self.utils.orderPointsData(data)
        obtained_nb_points_data = len(list(data.keys()))

        msg = ("The ascanct execution did not return any scan point.\n"
               "Checked using macro data.")
        self.assertTrue(len(list(data.keys())) > 0, msg)

        msg = ("The ascanct execution did not return the expected number of "
               "points.\nExpected " + str(expected_nb_points) + " points."
               "\nObtained " + str(obtained_nb_points_data) + " points."
               "\nChecked using macro data.")
        self.assertEqual(obtained_nb_points_data, expected_nb_points, msg)

        msg = ("Scan points are NOT in good order."
               "\nChecked using macro data.")
        self.assertTrue(order_points_data, msg)

    def check_stopped(self):
        self.assertStopped('Macro %s did not stop' % self.macro_name)
        for name in self.expchan_names + self.tg_names:
            channel = PyTango.DeviceProxy(name)
            desired_state = PyTango.DevState.ON
            state = channel.state()
            msg = 'element %s state after stop is %s (should be %s)' % \
                  (name, state, desired_state)
            self.assertEqual(state, desired_state, msg)

    def tearDown(self):
        RunStopMacroTestCase.tearDown(self)
        BaseMacroServerTestCase.tearDown(self)
        MeasSarTestTestCase.tearDown(self)


mg_config1 = {
    "_test_ct_ctrl_1": {
        "synchronizer": "_test_tg_1_1",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_1_1": {
                "index": 0
            },
        }
    },
}
mg_config2 = {
    "_test_ct_ctrl_1": {
        "synchronizer": "_test_tg_1_1",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_1_1": {
                "index": 0
            },
            "_test_ct_1_2": {
                "index": 1
            }
        }
    },
}
mg_config3 = {
    "_test_ct_ctrl_1": {
        "synchronizer": "software",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_1_1": {
                "index": 0
            },
        }
    },
    "_test_ct_ctrl_2": {
        "synchronizer": "_test_tg_1_1",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_2_1": {
                "index": 1
            },
        }
    }
}
mg_config4 = {
    "_test_ct_ctrl_1": {
        "synchronizer": "software",
        "synchronization": AcqSynchType.Trigger,
        "channels": {
            "_test_ct_1_1": {
                "index": 0
            },
            "_test_ct_1_2": {
                "index": 1
            }
        }
    }
}
ascanct_params_1 = ['_test_mt_1_1', '0', '10', '100', '0.1']


@testRun(meas_config=mg_config1, macro_params=ascanct_params_1,
         wait_timeout=30)
@testRun(meas_config=mg_config2, macro_params=ascanct_params_1,
         wait_timeout=30)
@testRun(meas_config=mg_config3, macro_params=ascanct_params_1,
         wait_timeout=30)
@testRun(meas_config=mg_config4, macro_params=ascanct_params_1,
         wait_timeout=30)
@testStop(meas_config=mg_config1, macro_params=ascanct_params_1,
          stop_delay=5, wait_timeout=20)
class AscanctTest(ScanctTest, unittest.TestCase):
    """Checks that ascanct works and generates the exact number of records
    by parsing the door output.

    .. todo:: check the macro data instead of the door output
    """
    macro_name = 'ascanct'

    def setUp(self):
        unittest.TestCase.setUp(self)
        ScanctTest.setUp(self)

    def macro_runs(self, meas_config, macro_params, wait_timeout=None):
        motors = [macro_params[0]]
        ScanctTest.configure_motors(self, motors)
        ScanctTest.configure_mntgrp(self, meas_config)
        # Run the ascanct
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=True, timeout=wait_timeout)
        self.assertFinished('Macro %s did not finish' % self.macro_name)

        expected_nb_points = int(macro_params[3]) + 1
        ScanctTest.check_using_output(self, expected_nb_points)
        ScanctTest.check_using_data(self, expected_nb_points)

    def macro_stops(self, meas_config, macro_params, wait_timeout=None,
                    stop_delay=0.1):
        motors = [macro_params[0]]
        ScanctTest.configure_motors(self, motors)
        ScanctTest.configure_mntgrp(self, meas_config)
        # Run the ascanct
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=False, timeout=wait_timeout)
        if stop_delay is not None:
            time.sleep(stop_delay)
        self.macro_executor.stop()
        self.macro_executor.wait(timeout=wait_timeout)
        ScanctTest.check_stopped(self)

    def tearDown(self):
        ScanctTest.tearDown(self)
        unittest.TestCase.tearDown(self)


a2scanct_params_1 = ['_test_mt_1_1', '0', '10', '_test_mt_1_2', '0', '20',
                     '100', '0.1']


@testRun(meas_config=mg_config1, macro_params=a2scanct_params_1,
         wait_timeout=30)
class A2scanctTest(ScanctTest, unittest.TestCase):
    """Checks that a2scanct works and generates the exact number of records
    by parsing the door output.

    .. todo:: check the macro data instead of the door output
    """
    macro_name = 'a2scanct'
    MOT1 = 0
    MOT2 = 3

    def setUp(self):
        unittest.TestCase.setUp(self)
        ScanctTest.setUp(self)

    def macro_runs(self, meas_config, macro_params, wait_timeout=None):
        motors = [macro_params[self.MOT1], macro_params[self.MOT2]]
        ScanctTest.configure_motors(self, motors)
        ScanctTest.configure_mntgrp(self, meas_config)
        # Run the ascanct
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=True, timeout=wait_timeout)
        self.assertFinished('Macro %s did not finish' % self.macro_name)

        expected_nb_points = int(macro_params[6]) + 1
        ScanctTest.check_using_output(self, expected_nb_points)
        ScanctTest.check_using_data(self, expected_nb_points)

    def macro_stops(self, meas_config, macro_params, wait_timeout=None,
                    stop_delay=0.1):
        motors = [macro_params[self.MOT1], macro_params[self.MOT2]]
        ScanctTest.configure_motors(self, motors)
        ScanctTest.configure_mntgrp(self, meas_config)
        # Run the ascanct
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=False, timeout=wait_timeout)
        if stop_delay is not None:
            time.sleep(stop_delay)
        self.macro_executor.stop()
        self.macro_executor.wait(timeout=wait_timeout)
        ScanctTest.check_stopped(self)

    def tearDown(self):
        ScanctTest.tearDown(self)
        unittest.TestCase.tearDown(self)


meshct_params_1 = ['_test_mt_1_1', '0', '10', '2', '_test_mt_1_2', '0', '20',
                   '2', '0.1']


@testRun(meas_config=mg_config1, macro_params=meshct_params_1,
         wait_timeout=30)
class MeshctTest(ScanctTest, unittest.TestCase):
    """Checks that meshct works and generates the exact number of records
    by parsing the door output.

    .. todo:: check the macro data instead of the door output
    """
    macro_name = 'meshct'
    MOT1 = 0
    MOT2 = 4
    INTERVALS_MOT1 = 3
    INTERVALS_MOT2 = 7

    def setUp(self):
        unittest.TestCase.setUp(self)
        ScanctTest.setUp(self)

    def macro_runs(self, meas_config, macro_params, wait_timeout=None):
        motors = [macro_params[self.MOT1], macro_params[self.MOT2]]
        ScanctTest.configure_motors(self, motors)
        ScanctTest.configure_mntgrp(self, meas_config)
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=True, timeout=wait_timeout)
        self.assertFinished('Macro %s did not finish' % self.macro_name)

        expected_nb_points = (int(macro_params[self.INTERVALS_MOT1]) + 1) * \
                             (int(macro_params[self.INTERVALS_MOT2]) + 1)
        ScanctTest.check_using_output(self, expected_nb_points)
        ScanctTest.check_using_data(self, expected_nb_points)

    def macro_stops(self, meas_config, macro_params, wait_timeout=None,
                    stop_delay=0.1):
        motors = [macro_params[self.MOT1], macro_params[self.MOT2]]
        ScanctTest.configure_motors(self, motors)
        ScanctTest.configure_mntgrp(self, meas_config)
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=False, timeout=wait_timeout)
        if stop_delay is not None:
            time.sleep(stop_delay)
        self.macro_executor.stop()
        self.macro_executor.wait(timeout=wait_timeout)
        ScanctTest.check_stopped(self)

    def tearDown(self):
        ScanctTest.tearDown(self)
        unittest.TestCase.tearDown(self)



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Experiment configuration related macros"""

__all__ = ["get_meas", "get_meas_conf", "set_meas", "set_meas_conf",
           "lssnap", "defsnap", "udefsnap"]

__docformat__ = 'restructuredtext'

from collections import OrderedDict

import taurus
from taurus.console import Alignment
from taurus.console.list import List

from sardana.pool import AcqSynchType
from sardana.macroserver.msexception import UnknownEnv
from sardana.taurus.core.tango.sardana import PlotType
from sardana.macroserver.macro import macro, Macro, Type, Optional


def sanitizer(values):
    return ["n/a" if v is None else v for v in values]


def plot_type_sanitizer(values):
    return [PlotType.whatis(v) for v in values]


def plot_axes_sanitizer(values):
    return ["n/a" if len(v) == 0 else v[0] for v in values]


def synchrtonization_sanitizer(values):
    return [AcqSynchType.whatis(v) for v in values]


def plot_axes_validator(value):
    value = value.lower()
    if value in ("idx", "<idx>"):
        value = ["<idx>"]
    elif value in ("mov", "<mov>"):
        value = ["<mov>"]
    return value


def bool_validator(value):
    in_value = value
    value = value.lower()
    if value in ['true', '1']:
        value = True
    elif value in ['false', '0']:
        value = False
    else:
        raise ValueError('{0} is not a boolean'.format(in_value))
    return value


def synchronization_validator(value):
    in_value = value
    value = value.lower()
    try:
        try:
            value = int(value)
        except ValueError:
            value = AcqSynchType.get(value.capitalize())
        else:
            value = AcqSynchType.get(value)
    except KeyError:
        raise ValueError("{0} is not a synchronization type".format(in_value))
    return value

# if sanitizers and validators evolve to sth too complicated refactor this
# to use classes
parameter_map = OrderedDict([
    ("enabled", ("Enabled", None, bool_validator)),
    ("output", ("Output", None, bool_validator)),
    ("plottype", ("PlotType", plot_type_sanitizer, None)),
    ("plotaxes", ("PlotAxes", plot_axes_sanitizer, plot_axes_validator)),
    ("timer", ("Timer", sanitizer, None)),
    ("monitor", ("Monitor", sanitizer, None)),
    ("synchronizer", ("Synchronizer", sanitizer, None)),
    ("synchronization",
        ("Synchronization", synchrtonization_sanitizer,
         synchronization_validator)),
    ("valuerefenabled", ("ValueRefEnabled", sanitizer, bool_validator)),
    ("valuerefpattern", ("ValueRefPattern", sanitizer, None))
])


simple_parameters = ("enabled", "plottype", "plotaxes")


@macro([
    ["detail", Type.String, Optional,
        "Detail level of parameters. If omitted or \"simple\" then "
        "get simple parameters. If \"all\" then get all parameters."],
    ["meas_grp", Type.MeasurementGroup, Optional, "Measurement group"]
])
def get_meas_conf(self, detail, meas_grp):
    """Print measurement group configuration in form of a table.

    Examples of usage:

    >>> get_meas_conf  # get <ActiveMntGrp> simple configuration
    >>> get_meas_conf all  # get <ActiveMntGrp> complete configuration
    >>> get_meas_conf simple mntgrp01  # get mntgrp01 simple configuration
    >>> get_meas_conf all mntgrp01  # get mntgrp01 complete configuration
    """
    if detail is None or detail == "simple":
        parameters = simple_parameters
    elif detail == "all":
        parameters = parameter_map.keys()
    else:
        raise ValueError("wrong detail level: {}".format(detail))
    if meas_grp is None:
        meas_grp = self.getEnv("ActiveMntGrp")
        self.print("ActiveMntGrp = {}".format(meas_grp))
        meas_grp = self.getMeasurementGroup(meas_grp)
    col_headers = ["Channel"]
    width = [-1]
    align = [Alignment.Right]
    cols = []
    for parameter in parameters:
        parameter, sanitizer, _ = parameter_map[parameter]
        getter = getattr(meas_grp, "get" + parameter)
        ret = getter()
        if len(cols) == 0:
            # add channel names as first column
            cols.append(ret.keys())
        values = ret.values()
        if sanitizer is not None:
            values = sanitizer(values)
        cols.append(values)
        col_headers.append(parameter)
        width.append(-1)
        align.append(Alignment.Right)
    out = List(col_headers, text_alignment=align, max_col_width=width)
    for row in zip(*cols):
        out.appendRow(row)
    for line in out.genOutput():
        self.output(line)


@macro([
    ["parameter", Type.String, None, "Parameter (case insensitive) to set."],
    ["value", Type.String, None, "Parameter value to set"],
    ["items", [
        ["item", Type.String, None, "Experimental channel/controller"],
        {"min": 0}], None,
        "Experimental channels (also external e.g. Tango attribute: "
        "sys/tg_test/1/ampli) or their controllers"
     ],
    ["meas_grp", Type.MeasurementGroup, Optional, "Measurement group"],
])
def set_meas_conf(self, parameter, value, items, meas_grp):
    """Set measurement group configuration parameter.

    Available configuration parameters and values:

    - **Enabled**: True/1 or False/0
    - **Output**: True/1 or False/0
    - **PlotType**: No, Spectrum or Image
    - **PlotAxes**: idx, mov or a <channel name> e.g. ct01
      (for image use "|" as separator of axes e.g. idx|ct01)
    - **Timer**: <channel name> e.g. ct01
    - **Monitor**: <channel name> e.g. ct01
    - **Synchronizer**: software or <trigger/gate name> e.g. tg01
    - **Synchronization**: 0/Trigger, 1/Gate or 2/Start
    - **ValueRefEnabled** - True/1 or False/0
    - **ValueRefPattern** - URI e.g. file:///tmp/img_{index}.tiff

    Examples of usage:

    >>> # enable all channels in <ActiveMntGrp>
    >>> set_meas_conf enabled True
    >>> # enable spectrum plotting for ct01 on <ActiveMntGrp>
    >>> set_meas_conf plottype spectrum ct01
    >>> # set plot x-axis to <moveable> for all channels of <ActiveMntGrp>
    >>> set_meas_conf plotaxes mov
    >>> # enable spectrum plotting for all mntgrp01
    >>> set_meas_conf plottype spectrum [] mntgrp01
    """
    try:
        parameter, _, validator = parameter_map[parameter.lower()]
    except KeyError:
        raise ValueError("wrong parameter: {}".format(parameter))
    if meas_grp is None:
        meas_grp = self.getEnv("ActiveMntGrp")
        self.print("ActiveMntGrp = {}".format(meas_grp))
        meas_grp = self.getMeasurementGroup(meas_grp)
    setter = getattr(meas_grp, "set" + parameter)
    if validator is not None:
        value = validator(value)
    setter(value, *items)


@macro([
    ["meas_grp", Type.MeasurementGroup, None,
        "Measurement group to activate"],
    ["macro", Type.Macro, Optional,
        "Activate measurement group on this macro."
        "If omitted activate on all."],
])
def set_meas(self, meas_grp, macro):
    """Activate measurement group.

    It sets the ActiveMntGrp environment variable.
    """
    var = "ActiveMntGrp"
    if macro is not None:
        var = "{}.{}".format(macro.name, var)
    self.setEnv(var, meas_grp.getName())


@macro([
    ["macro", Type.Macro, Optional,
        "Get active measurement group set for this macro."
        "If omitted get the one set for all."],
])
def get_meas(self, macro):
    """Activate measurement group.

    It gets the ActiveMntGrp environment variable.
    """
    if macro is not None:
        macro_name = macro.name
    else:
        macro_name = None
    var = "ActiveMntGrp"
    try:
        value = self.getEnv(var, macro_name=macro_name)
    except UnknownEnv:
        self.warning("Active measurement group is not set. "
                     "Hint: use `set_meas` macro to set it.")
    else:
        self.print("{} = {}".format(var, value))


class lssnap(Macro):
    """List pre-scan snapshot group.
    """

    def run(self):
        try:
            snapshot_items = self.getEnv("PreScanSnapshot")
        except UnknownEnv:
            self.output("No pre-scan snapshot")
            return
        out = List(['Snap item', 'Snap item full name'])
        for full_name, label in snapshot_items:
            out.appendRow([label, full_name])
        for line in out.genOutput():
            self.output(line)


class defsnap(Macro):
    """Define snapshot group item(s).

    Accepts:

    - Pool moveables: motor, pseudo motor
    - Pool experimental channels: counter/timer, 0D, 1D, 2D, pseudo counter
    - Taurus attributes
    """

    param_def = [
        ["snap_names", [[
            "name", Type.String, None, "Name of an item to be added to the "
                                       "pre-scan snapshot group"]],
            None,
            "Items to be added to the pre-scan snapshot group"],
    ]

    def run(self, snap_names):

        def get_item_info(item):
            if isinstance(item, taurus.core.TaurusAttribute):
                return item.fullname, item.label
            else:
                return item.full_name, item.name
        try:
            snap_items = self.getEnv("PreScanSnapshot")
        except UnknownEnv:
            snap_items = []
        snap_full_names = [item[0] for item in snap_items]
        new_snap_items = []
        for name in snap_names:
            obj = self.getObj(name)
            if obj is None:
                try:
                    obj = taurus.Attribute(name)
                except taurus.TaurusException:
                    raise ValueError("item is neither Pool element not "
                                     "Taurus attribute")
            elif obj.type == "MotorGroup":
                raise ValueError("MotorGroup item type is not accepted")
            new_full_name, new_label = get_item_info(obj)
            if new_full_name in snap_full_names:
                msg = "{} already in pre-scan snapshot".format(name)
                raise ValueError(msg)
            new_snap_items.append((new_full_name, new_label))
        self.setEnv("PreScanSnapshot", snap_items + new_snap_items)


class udefsnap(Macro):
    """Undefine snapshot group item(s). Without arguments undefine all.
    """

    param_def = [
        ["snap_names", [[
            "name", Type.String, None, "Name of an item to be removed "
                                       "from the pre-scan snapshot group",
            ], {"min": 0}],
         None,
         "Items to be remove from the pre-scan snapshot group"],
    ]

    def run(self, snap_names):
        if len(snap_names) == 0:
            self.unsetEnv("PreScanSnapshot")
            return
        try:
            snap_items = self.getEnv("PreScanSnapshot")
        except UnknownEnv:
            raise RuntimeError("no pre-scan snapshot defined")
        snap_full_names = {}
        for i, item in enumerate(snap_items):
            snap_full_names[item[0]] = i
        for name in snap_names:
            obj = self.getObj(name)
            if obj is None:
                try:
                    obj = taurus.Attribute(name)
                except taurus.TaurusException:
                    raise ValueError("item is neither Pool element not "
                                     "Taurus attribute")
            elif obj.type == "MotorGroup":
                raise ValueError("MotorGroup item type is not accepted")
            rm_full_name = obj.fullname
            if rm_full_name not in snap_full_names.keys():
                msg = "{} not in pre-scan snapshot".format(name)
                raise ValueError(msg)
            i = snap_full_names[rm_full_name]
            snap_items.pop(i)
        self.setEnv("PreScanSnapshot", snap_items)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains tests for HDF5 recorders."""

import os
import tempfile
from datetime import datetime

import h5py
import numpy
from unittest import TestCase

from sardana.macroserver.scan import ColumnDesc
from sardana.macroserver.recorders.h5storage import NXscanH5_FileRecorder

COL1_NAME = "col1"


class RecordList(dict):

    def __init__(self, env):
        self._env = env

    def getEnviron(self):
        return self._env


class Record(object):

    def __init__(self, data, recordno=0):
        self.data = data
        self.recordno = recordno


class TestNXscanH5_FileRecorder(TestCase):

    def setUp(self):
        self.dir_name = tempfile.gettempdir()
        self.path = os.path.join(self.dir_name, "test.h5")
        try:
            os.remove(self.path)  # remove file just in case
        except OSError:
            pass

        self.env = {
            "serialno": 0,
            "starttime": None,
            "title": "test",
            "user": "user",
            "datadesc": None,
            "endtime": None
        }
        self.record_list = RecordList(self.env)

    def test_dtype_float64(self):
        """Test creation of dataset with float64 data type"""
        nb_records = 1
        # create description of channel data
        data_desc = [
            ColumnDesc(name=COL1_NAME, label=COL1_NAME, dtype="float64",
                       shape=tuple())
        ]
        self.env["datadesc"] = data_desc

        # simulate sardana scan
        recorder = NXscanH5_FileRecorder(filename=self.path)
        self.env["starttime"] = datetime.now()
        recorder._startRecordList(self.record_list)
        for i in range(nb_records):
            record = Record({COL1_NAME: 0.1}, i)
            recorder._writeRecord(record)
        self.env["endtime"] = datetime.now()
        recorder._endRecordList(self.record_list)

        # assert if reading datasets from the sardana file access to the
        # dataset of the partial files
        file_ = h5py.File(self.path)
        for i in range(nb_records):
            expected_data = 0.1
            data = file_["entry0"]["measurement"][COL1_NAME][i]
            msg = "data does not match"
            self.assertEqual(data, expected_data, msg)

    def test_value_ref(self):
        """Test creation of dataset with str data type"""
        nb_records = 1
        # create description of channel data
        data_desc = [
            ColumnDesc(name=COL1_NAME, label=COL1_NAME, dtype="float64",
                       shape=(1024, 1024), value_ref_enabled=True)
        ]
        self.env["datadesc"] = data_desc

        # simulate sardana scan
        recorder = NXscanH5_FileRecorder(filename=self.path)
        self.env["starttime"] = datetime.now()
        recorder._startRecordList(self.record_list)
        for i in range(nb_records):
            record = Record({COL1_NAME: "file:///tmp/test.edf"}, i)
            recorder._writeRecord(record)
        self.env["endtime"] = datetime.now()
        recorder._endRecordList(self.record_list)

        # assert if reading datasets from the sardana file access to the
        # dataset of the partial files
        file_ = h5py.File(self.path)
        for i in range(nb_records):
            expected_data = "file:///tmp/test.edf"
            data = file_["entry0"]["measurement"][COL1_NAME][i]
            msg = "data does not match"
            self.assertEqual(data, expected_data, msg)

    def test_VDS(self):
        """Test creation of VDS when channel reports URIs (str) of h5file
        scheme in a simulated sardana scan (3 points).
        """
        try:
            h5py.VirtualLayout
        except AttributeError:
            self.skipTest("VDS not available in this version of h5py")
        nb_records = 3
        # create partial files
        part_file_name_pattern = "test_vds_part{0}.h5"
        part_file_paths = []
        for i in range(nb_records):
            path = os.path.join(self.dir_name,
                                part_file_name_pattern.format(i))
            part_file_paths.append(path)
            part_file = h5py.File(path, "w")
            img = numpy.array([[i, i], [i, i]])
            dataset = "dataset"
            part_file.create_dataset(dataset, data=img)
            part_file.flush()
            part_file.close()
        try:
            # create description of channel data
            data_desc = [
                ColumnDesc(name=COL1_NAME, label=COL1_NAME, dtype="float64",
                           shape=(2, 2), value_ref_enabled=True)
            ]
            self.env["datadesc"] = data_desc

            # simulate sardana scan
            recorder = NXscanH5_FileRecorder(filename=self.path)
            self.env["starttime"] = datetime.now()
            recorder._startRecordList(self.record_list)
            for i in range(nb_records):
                ref = "h5file://" + part_file_paths[i] + "::" + dataset
                record = Record({COL1_NAME: ref}, i)
                recorder._writeRecord(record)
            self.env["endtime"] = datetime.now()
            recorder._endRecordList(self.record_list)

            # assert if reading datasets from the sardana file access to the
            # dataset of the partial files
            file_ = h5py.File(self.path)
            for i in range(nb_records):
                expected_img = numpy.array([[i, i], [i, i]])
                img = file_["entry0"]["measurement"][COL1_NAME][i]
                msg = "VDS extracted image does not match"
                # TODO: check if this assert works well
                numpy.testing.assert_array_equal(img, expected_img, msg)
        finally:
            # remove partial files
            for path in part_file_paths:
                os.remove(path)

    def tearDown(self):
        try:
            os.remove(self.path)
        except OSError:
            pass



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import unittest

from sardana.pool.poolmotion import PoolMotion
from sardana.sardanadefs import State
from sardana.pool.test import (FakePool, createPoolController,
                               createPoolMotor, dummyPoolMotorCtrlConf01,
                               dummyMotorConf01, dummyMotorConf02)


class PoolMotionTestCase(unittest.TestCase):
    """Unittest of PoolMotion class"""

    def setUp(self):
        """Create a Controller, and Motor objects from dummy configurations """
        unittest.TestCase.setUp(self)
        try:
            from mock import Mock
        except ImportError:
            self.skipTest("mock module is not available")
        pool = FakePool()
        dummy_mot_ctrl = createPoolController(pool, dummyPoolMotorCtrlConf01)
        self.dummy_mot = createPoolMotor(pool, dummy_mot_ctrl,
                                         dummyMotorConf01)
        self.dummy_mot2 = createPoolMotor(pool, dummy_mot_ctrl,
                                          dummyMotorConf02)
        dummy_mot_ctrl.add_element(self.dummy_mot)
        pool.add_element(dummy_mot_ctrl)
        pool.add_element(self.dummy_mot)
        pool.add_element(self.dummy_mot2)

        # {moveable: (position, dial_position,
        #             do_backlash, backlash, instability_time=None)}
        self.items = {self.dummy_mot: (0, 0, False, 0),
                      self.dummy_mot2: (0, 0, False, 0)}
        # Create mock and define its functions
        ctrl_methods = ['PreStartAll', 'StartAll', 'PreStartOne', 'StartOne',
                        'PreStateAll', 'StateAll', 'PreStateOne', 'StateOne',
                        'PreReadAll', 'PreReadOne', 'ReadOne', 'ReadAll',
                        'PreStopAll', 'StopAll', 'PreStopOne', 'StopOne',
                        'PreAbortAll', 'AbortAll', 'PreAbortOne', 'AbortOne']
        self.mock_mot_ctrl = Mock(spec=ctrl_methods)
        self.mock_mot_ctrl.StateOne.return_value = (State.Moving, 'moving')

        dummy_mot_ctrl.ctrl = self.mock_mot_ctrl
        self.motionaction = PoolMotion(self.dummy_mot)
        self.motionaction.add_element(self.dummy_mot)
        self.motionaction.add_element(self.dummy_mot2)

    def stopMotion(self):
        """Method used to change the controller (mock) state"""
        self.dummy_mot.stop()

    def test_stop(self):
        """Verify motion stop call chain."""
        from mock import call
        args = ()
        kwargs = {'items': self.items}
        # starting action
        self.motionaction.start_action(*args, **kwargs)
        # stopping the motion
        # self.stopMotion()
        args = ()
        kwargs = {'items': self.items}
        self.motionaction.stop_action(*args, **kwargs)

        # verifying that the stop has called all the controller methods chain
        self.mock_mot_ctrl.assert_has_calls([call.PreStopAll(),
                                             call.PreStopOne(1,),
                                             call.StopOne(1,),
                                             call.PreStopOne(2,),
                                             call.StopOne(2,),
                                             call.StopAll()])

    def test_abort(self):
        """Verify motion abort call chain."""
        from mock import call
        args = ()
        kwargs = {'items': self.items}
        # starting action
        self.motionaction.start_action(*args, **kwargs)
        args = ()
        kwargs = {'items': self.items}
        self.motionaction.abort_action(*args, **kwargs)

        # verifying that the abort has called all the controller methods chain
        self.mock_mot_ctrl.assert_has_calls([call.PreAbortAll(),
                                             call.PreAbortOne(1,),
                                             call.AbortOne(1,),
                                             call.PreAbortOne(2,),
                                             call.AbortOne(2,),
                                             call.AbortAll()])

    def tearDown(self):
        self.motionaction = None
        self.mock_mot_ctrl = None
        self.cfg = None
        self.dummy_mot = None
        unittest.TestCase.tearDown(self)



import sys
import os

from PyQt4 import QtGui, QtCore, Qt

import taurus
import wiz


def get_default_tango_host():
    tg = os.environ.get('TANGO_HOST', '')
    if not tg or tg.count(':') != 1:
        return '', ''
    return tg.split(':')


class SelectTangoHostBasePage(wiz.SardanaBasePage):

    def __init__(self, parent=None):
        wiz.SardanaBasePage.__init__(self, parent)

        self.setSubTitle('Please select the Tango Host')

        panel = self.getPanelWidget()
        layout = QtGui.QFormLayout()
        self.hostEdit = QtGui.QLineEdit()
        self.portEdit = QtGui.QLineEdit()

        layout.addRow("&Host", self.hostEdit)
        layout.addRow("&Port", self.portEdit)

        host, port = get_default_tango_host()

        self.hostEdit.setText(host)
        self.portEdit.setText(port)

        panel.setLayout(layout)

        self['db'] = self._getDatabase

        self.connect(self.hostEdit,
                     QtCore.SIGNAL('textEdited(const QString &)'),
                     QtCore.SIGNAL('completeChanged()'))
        self.connect(self.portEdit,
                     QtCore.SIGNAL('textEdited(const QString &)'),
                     QtCore.SIGNAL('completeChanged()'))

    def _getDatabase(self):
        host = str(self.hostEdit.text()).lower()
        port = str(self.portEdit.text())
        return taurus.Database("%s:%s" % (host, port))

    def db(self):
        return self.wizard()['db']

    def isComplete(self):
        try:
            db = self.db()
            if not db is None:
                self.setStatus('')
                return True
        except Exception, e:
            pass
        self.setStatus('Invalid database')
        return False


def t1():
    app = QtGui.QApplication([])
    QtCore.QResource.registerResource(wiz.get_resources())
    w = wiz.SardanaBaseWizard()
    intro = wiz.SardanaIntroBasePage()
    tg_host_page = SelectTangoHostBasePage()
    w.addPage(intro)
    w.addPage(tg_host_page)
    w.show()
    sys.exit(app.exec_())


def main():
    t1()

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import sys
import types
from lxml import etree

SAR_NS = 'http://sardana.cells.es/client/framework/config'


def transform(f):
    if isinstance(f, str):
        doc = etree.parse(f)
    else:
        doc = f

    root = doc.getroot()

    if (root.nsmap.get(None) == SAR_NS) or root.tag.endswith("Sardana"):
        return doc

    # it is either a EXCEL XML or a Flat ODS
    if root.nsmap.get('ss'):
        import xls_to_sar
        t = xls_to_sar.transform
    else:
        import fods_to_sar
        t = fods_to_sar.transform
    return t(doc)


def main():
    filename = sys.argv[1]
    t = transform(filename)
    print(etree.tostring(t, pretty_print=True))

if __name__ == "__main__":
    main()



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains the definition of a I/I0 pseudo counter controller
for the Sardana Device Pool"""

__all__ = ["IoverI0"]

__docformat__ = 'restructuredtext'

from sardana.pool.controller import PseudoCounterController


class IoverI0(PseudoCounterController):
    """ A simple pseudo counter which receives two counter values (I and I0)
        and returns I/I0"""

    gender = "IoverI0"
    model = "Default I/I0"
    organization = "Sardana team"

    counter_roles = "I", "I0"

    def Calc(self, axis, counter_values):
        i, i0 = counter_values
        try:
            i = i / i0
        except ZeroDivisionError:
            pass
        return i
